#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from mitosheet.state import DATAFRAME_SOURCE_DUPLICATED, DATAFRAME_SOURCE_IMPORTED, DATAFRAME_SOURCE_MERGED, DATAFRAME_SOURCE_PASSED, DATAFRAME_SOURCE_PIVOTED
from mitosheet.user.schemas import USER_JSON_DEFAULT
from mitosheet.step_performers import STEP_PERFORMERS
from mitosheet.step_performers.pivot import PIVOT_AGGREGATION_TYPES
from mitosheet.step_performers.sort import ASCENDING, DESCENDING
from mitosheet.step_performers.filter import BOOLEAN_FILTER_CONDITIONS, FC_EMPTY, FC_NOT_EMPTY, STRING_FILTER_CONDITIONS, NUMBER_FILTER_CONDITIONS, DATETIME_FILTER_CONDITIONS
from mitosheet.sheet_functions.types.utils import BOOLEAN_SERIES, DATETIME_SERIES, NUMBER_SERIES, STRING_SERIES, TIMEDELTA_SERIES
from mitosheet.data_in_mito import DataTypeInMito

def get_enum_from_ts_file(path, enum_name):
    """
    Helper function that retuns an enum from a TS
    file, parsed as a JSON object. 

    NOTE: this only works for enums that are spread
    out across mulitple lines :-) 
    """
    with open(path, 'r+') as f:
        all_code_lines = f.readlines()
    
    # First, we find where the enum starts and ends
    start_line_index = None
    end_line_index = None
    for index, code_line in enumerate(all_code_lines):
        if f'enum {enum_name}' in code_line:
            start_line_index = index
        elif start_line_index is not None and '}' in code_line:
            end_line_index = index
            break
    
    if start_line_index is None:
        raise Exception(f'{enum_name} is not in {path}, did you move it or rename it?')
    
    # Then, for each item in the enum, we split it into key value pairs
    enum_item_code_lines = all_code_lines[start_line_index + 1:end_line_index]
    enum_items = {}
    for enum_item_code_line in enum_item_code_lines:
        enum_item_code_line = enum_item_code_line.strip()
        # Remove trailing commas
        if enum_item_code_line.endswith(','):
            enum_item_code_line = enum_item_code_line[:-1]
        
        # Split into keys and values
        [name, value] = enum_item_code_line.split(' = ')
        # If the value is a string and has quotes around it, 
        # take them off
        if value.startswith('\'') or value.startswith('\"'):
            value = value[1:]
        if value.endswith('\'') or value.endswith('\"'):
            value = value[:-1]

        enum_items[name] = value

    return enum_items

def test_column_types_match():
    column_mito_type = get_enum_from_ts_file('./src/components/Mito.tsx', 'ColumnMitoType')
    assert len(column_mito_type) == 5
    assert column_mito_type['BOOLEAN_SERIES'] == BOOLEAN_SERIES
    assert column_mito_type['STRING_SERIES'] == STRING_SERIES
    assert column_mito_type['NUMBER_SERIES'] == NUMBER_SERIES
    assert column_mito_type['DATETIME_SERIES'] == DATETIME_SERIES
    assert column_mito_type['TIMEDELTA_SERIES'] == TIMEDELTA_SERIES


def test_date_type_in_mito_match():
    data_type_in_mito = get_enum_from_ts_file('./src/components/Mito.tsx', 'DataTypeInMito')
    assert len(data_type_in_mito) == 4

    assert data_type_in_mito['NONE'] == DataTypeInMito.NONE.value
    assert data_type_in_mito['PROVIDED'] == DataTypeInMito.PROVIDED.value
    assert data_type_in_mito['TUTORIAL'] == DataTypeInMito.TUTORIAL.value
    assert data_type_in_mito['PERSONAL'] == DataTypeInMito.PERSONAL.value


def test_sort_direction_match():
    sort_direcion = get_enum_from_ts_file('./src/components/taskpanes/ControlPanel/SortCard.tsx', 'SortDirection')
    # NOTE: the extra element here is no sort (on the frontend)
    assert len(sort_direcion) == 3

    assert sort_direcion['ASCENDING'] == ASCENDING
    assert sort_direcion['DESCENDING'] == DESCENDING


def test_filter_conditions_match():
    boolean_filter_conditions = get_enum_from_ts_file('./src/components/taskpanes/ControlPanel/filter/filterTypes.tsx', 'BooleanFilterCondition')
    assert len(boolean_filter_conditions) == len(BOOLEAN_FILTER_CONDITIONS) + 2
    assert set(BOOLEAN_FILTER_CONDITIONS).union(set([FC_EMPTY, FC_NOT_EMPTY])) == set(boolean_filter_conditions.values())
    
    number_filter_conditions = get_enum_from_ts_file('./src/components/taskpanes/ControlPanel/filter/filterTypes.tsx', 'NumberFilterCondition')
    assert len(number_filter_conditions) == len(NUMBER_FILTER_CONDITIONS) + 2
    assert set(NUMBER_FILTER_CONDITIONS).union(set([FC_EMPTY, FC_NOT_EMPTY])) == set(number_filter_conditions.values())
    
    string_filter_conditions = get_enum_from_ts_file('./src/components/taskpanes/ControlPanel/filter/filterTypes.tsx', 'StringFilterCondition')
    assert len(string_filter_conditions) == len(STRING_FILTER_CONDITIONS) + 2
    assert set(string_filter_conditions.values()) == set(STRING_FILTER_CONDITIONS).union(set([FC_EMPTY, FC_NOT_EMPTY]))

    datetime_filter_conditions = get_enum_from_ts_file('./src/components/taskpanes/ControlPanel/filter/filterTypes.tsx', 'DatetimeFilterCondition')
    assert len(datetime_filter_conditions) == len(DATETIME_FILTER_CONDITIONS) + 2
    assert set(DATETIME_FILTER_CONDITIONS).union(set([FC_EMPTY, FC_NOT_EMPTY])) == set(datetime_filter_conditions.values())


def test_pivot_aggregation_functions_match():
    aggregation_types = get_enum_from_ts_file('./src/components/taskpanes/PivotTable/PivotTaskpane.tsx', 'AggregationType')
    assert set(aggregation_types.values()) == set(PIVOT_AGGREGATION_TYPES)


def test_step_types_match():
    step_types = get_enum_from_ts_file('./src/types/StepTypes.ts', 'StepType')
    assert set(step_types.values()) == set([step_performers.step_type() for step_performers in STEP_PERFORMERS] + ['initialize'])

def test_df_sources_matche():
    df_sources = get_enum_from_ts_file('./src/types/DFSources.ts', 'DFSource')
    assert set(df_sources.values()) == set([DATAFRAME_SOURCE_PASSED, DATAFRAME_SOURCE_IMPORTED, DATAFRAME_SOURCE_PIVOTED, DATAFRAME_SOURCE_MERGED, DATAFRAME_SOURCE_DUPLICATED])


def test_user_json_fields_match():
    user_json_fields = get_enum_from_ts_file('./src/api.tsx', 'UserJsonFields')
    assert set(user_json_fields.values()) == set(USER_JSON_DEFAULT.keys())


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import copy
from typing import Any, Dict, List, Optional
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.utils import make_valid_header
from mitosheet.step_performers.dataframe_steps.dataframe_duplicate import get_first_unused_name

class DataframeRenameStepPerformer(StepPerformer):
    """"
    A rename dataframe step changes the name of a specific dataframe
    at a specific index.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'dataframe_rename'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Renamed a Dataframe'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'dataframe_rename_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        sheet_index = params['sheet_index']
        old_dataframe_name = prev_state.df_names[sheet_index]
        params['old_dataframe_name'] = old_dataframe_name
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        old_dataframe_name: str,
        new_dataframe_name: str,
        **params
    ) -> State:
        # Bail early, if there is no change
        if old_dataframe_name == new_dataframe_name:
            return

        # Create a new step and save the parameters
        post_state = copy(prev_state)

        post_state.df_names[sheet_index] = get_valid_dataframe_name(post_state.df_names, new_dataframe_name)

        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        old_dataframe_name: str,
        new_dataframe_name: str
    ) -> List[str]:
        if old_dataframe_name == new_dataframe_name:
            return []
        return [f'{post_state.df_names[sheet_index]} = {old_dataframe_name}']

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        old_dataframe_name: str,
        new_dataframe_name: str,
        df_names=None,
        **params
    ) -> str:
        return f'Renamed {old_dataframe_name} to {new_dataframe_name}'


def get_valid_python_variable_name(suggested_dataframe_name):
    """
    Turns the suggested dataframe name into a valid Python identifier.
    """
    # See: https://www.w3schools.com/python/ref_string_isidentifier.asp
    if suggested_dataframe_name.isidentifier():
        return suggested_dataframe_name

    # First, we try and just fill in the non valid characters (like a column header)
    valid_header_version = make_valid_header(suggested_dataframe_name)
    if valid_header_version.isidentifier():
        return valid_header_version

    # Finially prefix with df, so that it doesn't start with a number (as vars cannot)
    return f'df_{valid_header_version}'


def get_valid_dataframe_name(df_names, suggested_dataframe_name):
    """
    Given an attempt at a dataframe name, makes sure that is a valid
    dataframe name (e.g. it is a valid Python identifier, and it is
    not a duplicate).
    """
    valid_identifier = get_valid_python_variable_name(suggested_dataframe_name)
    return get_first_unused_name(df_names, valid_identifier)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import copy
from mitosheet.state import State
from typing import Any, Dict, List, Optional
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.mito_analytics import log
from mitosheet.errors import make_invalid_column_type_change_error
from mitosheet.sheet_functions.types.to_number_series import to_number_series
from mitosheet.sheet_functions.types.to_boolean_series import to_boolean_series
from mitosheet.sheet_functions.types.utils import (
    get_datetime_format, get_mito_type, is_bool_dtype, is_float_dtype, is_int_dtype,
    is_string_dtype, is_datetime_dtype, is_timedelta_dtype
) 
import pandas as pd

 
class ChangeColumnDtypeStepPerformer(StepPerformer):
    """"
    A step that allows changing the dtype of a column to a different
    dtype.

    Currently, supports: 'bool', 'int', 'float', 'str', 'datetime', 'timedelta'
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'change_column_dtype'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Change column dtype'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'change_column_dtype_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        params['old_dtype'] = str(prev_state.dfs[params['sheet_index']][params['column_header']].dtype)
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        column_header: str,
        old_dtype: str,
        new_dtype: str,
        **params
    ) -> State:
        # Create the post state
        post_state = copy(prev_state)
        
        column: pd.Series = prev_state.dfs[sheet_index][column_header]
        new_column = column
        
        # How we handle the type conversion depends on what type it is
        try:
            if is_bool_dtype(old_dtype):
                if is_bool_dtype(new_dtype):
                    pass
                elif is_int_dtype(new_dtype):
                    new_column = new_column.astype('int')
                elif is_float_dtype(new_dtype):
                    new_column = column.astype('float')
                elif is_string_dtype(new_dtype):
                    new_column = column.astype('str')
                elif is_datetime_dtype(new_dtype):
                    raise make_invalid_column_type_change_error(
                        column_header,
                        old_dtype,
                        new_dtype
                    )
                elif is_timedelta_dtype(new_dtype):
                    raise make_invalid_column_type_change_error(
                        column_header,
                        old_dtype,
                        new_dtype
                    )
            if is_int_dtype(old_dtype):
                if is_bool_dtype(new_dtype):
                    new_column = column.astype('bool')
                elif is_int_dtype(new_dtype):
                    pass
                elif is_float_dtype(new_dtype):
                    new_column = column.astype('float')
                elif is_string_dtype(new_dtype):
                    new_column = column.astype('str')
                elif is_datetime_dtype(new_dtype):
                    new_column = pd.to_datetime(
                        column, 
                        unit='s',
                        errors='coerce'
                    )
                elif is_timedelta_dtype(new_dtype):
                    new_column = pd.to_timedelta(
                        column, 
                        unit='s',
                        errors='coerce'
                    )
            elif is_float_dtype(old_dtype):
                if is_bool_dtype(new_dtype):
                    new_column = column.astype('bool')
                elif is_int_dtype(new_dtype):
                    new_column = column.astype('int')
                elif is_float_dtype(new_dtype):
                    pass
                elif is_string_dtype(new_dtype):
                    new_column = column.astype('str')
                elif is_datetime_dtype(new_dtype):
                    new_column = pd.to_datetime(
                        column, 
                        unit='s',
                        errors='coerce'
                    )
                elif is_timedelta_dtype(new_dtype):
                    new_column = pd.to_timedelta(
                        column, 
                        unit='s',
                        errors='coerce'
                    )
            elif is_string_dtype(old_dtype):
                if is_bool_dtype(new_dtype):
                    new_column = to_boolean_series(new_column)
                elif is_int_dtype(new_dtype):
                    new_column = to_number_series(column).astype('int')
                elif is_float_dtype(new_dtype):
                    new_column = to_number_series(column)
                elif is_string_dtype(new_dtype):
                    pass
                elif is_datetime_dtype(new_dtype):
                    # Guess the datetime format to the best of Pandas abilities
                    datetime_format = get_datetime_format(column)
                    # If it's None, then infer_datetime_format is enough to figure it out
                    if datetime_format is not None:
                        new_column = pd.to_datetime(
                            column,
                            format=datetime_format,
                            errors='coerce'
                        )
                    else:
                        new_column = pd.to_datetime(
                            column,
                            infer_datetime_format=True,
                            errors='coerce'
                        )
                elif is_timedelta_dtype(new_dtype):
                    new_column = pd.to_timedelta(
                        column,
                        errors='coerce'
                    )
            elif is_datetime_dtype(old_dtype):
                if is_bool_dtype(new_dtype):
                    new_column = ~column.isnull()
                elif is_int_dtype(new_dtype):
                    new_column = column.astype('int') / 10**9
                elif is_float_dtype(new_dtype):
                    # For some reason, we have to do all the conversions at once
                    new_column = column.astype('int').astype('float') / 10**9
                elif is_string_dtype(new_dtype):
                    # NOTE: this is the same conversion that we send to the frontend
                    new_column = column.dt.strftime('%Y-%m-%d %X')
                elif is_datetime_dtype(new_dtype):
                    pass
                elif is_timedelta_dtype(new_dtype):
                    raise make_invalid_column_type_change_error(
                        column_header,
                        old_dtype,
                        new_dtype
                    )
            elif is_timedelta_dtype(old_dtype):
                if is_bool_dtype(new_dtype):
                    new_column = ~column.isnull()
                elif is_int_dtype(new_dtype):
                    new_column = column.dt.total_seconds().astype('int')
                elif is_float_dtype(new_dtype):
                    new_column = column.dt.total_seconds()
                elif is_string_dtype(new_dtype):
                    new_column = column.astype('str')
                elif is_datetime_dtype(new_dtype):
                    raise make_invalid_column_type_change_error(
                        column_header,
                        old_dtype,
                        new_dtype
                    )
                elif is_timedelta_dtype(new_dtype):
                    pass

            # We update the column, as well as the type of the column
            post_state.dfs[sheet_index][column_header] = new_column
            post_state.column_type[sheet_index][column_header] = get_mito_type(new_column)
            
            # We check if there are formula columns that rely on this type, and log if
            # so, so that know if this is an issue folks run into
            if len(post_state.column_evaluation_graph[sheet_index][column_header]) > 0:
                log(
                    'change_column_dtype_column_has_dependents',
                    {
                        'column_header': column_header
                    }
                )

            return post_state, None
        except:
            raise make_invalid_column_type_change_error(
                column_header,
                old_dtype,
                new_dtype
            )
        

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        column_header: str,
        old_dtype: str,
        new_dtype: str
    ) -> List[str]:
        df_name = post_state.df_names[sheet_index]
        column = prev_state.dfs[sheet_index][column_header]

        conversion_code = f'{df_name}[\'{column_header}\']'
        if is_bool_dtype(old_dtype):
            if is_bool_dtype(new_dtype):
                pass
            elif is_int_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'int\')'
            elif is_float_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'float\')'
            elif is_string_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'str\')'
            elif is_datetime_dtype(new_dtype):
                raise make_invalid_column_type_change_error(
                    column_header,
                    old_dtype,
                    new_dtype
                )
            elif is_timedelta_dtype(new_dtype):
                raise make_invalid_column_type_change_error(
                    column_header,
                    old_dtype,
                    new_dtype
                )
        elif is_int_dtype(old_dtype):
            if is_bool_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'bool\')'
            elif is_int_dtype(new_dtype):
                pass
            elif is_float_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'float\')'
            elif is_string_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'str\')'
            elif is_datetime_dtype(new_dtype):
                conversion_code = f'pd.to_datetime({df_name}[\'{column_header}\'], unit=\'s\', errors=\'coerce\')'
            elif is_timedelta_dtype(new_dtype):
                conversion_code = f'pd.to_timedelta({df_name}[\'{column_header}\'], unit=\'s\', errors=\'coerce\')'
        elif is_float_dtype(old_dtype):
            if is_bool_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'bool\')'
            elif is_int_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'int\')'
            elif is_float_dtype(new_dtype):
                pass
            elif is_string_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'str\')'
            elif is_datetime_dtype(new_dtype):
                conversion_code = f'pd.to_datetime({df_name}[\'{column_header}\'], unit=\'s\', errors=\'coerce\')'
            elif is_timedelta_dtype(new_dtype):
                conversion_code = f'pd.to_timedelta({df_name}[\'{column_header}\'], unit=\'s\', errors=\'coerce\')'
        elif is_string_dtype(old_dtype):
            if is_bool_dtype(new_dtype):
                conversion_code = f'to_boolean_series({df_name}[\'{column_header}\'])'
            elif is_int_dtype(new_dtype):
                conversion_code = f'to_number_series({df_name}[\'{column_header}\']).astype(\'int\')'
            elif is_float_dtype(new_dtype):
                conversion_code = f'to_number_series({df_name}[\'{column_header}\'])'
            elif is_string_dtype(new_dtype):
                pass
            elif is_datetime_dtype(new_dtype):
                # Guess the datetime format to the best of Pandas abilities
                datetime_format = get_datetime_format(column)
                if datetime_format is not None:
                    conversion_code = f'pd.to_datetime({df_name}[\'{column_header}\'], format=\'{datetime_format}\', errors=\'coerce\')'
                else:
                    conversion_code = f'pd.to_datetime({df_name}[\'{column_header}\'], infer_datetime_format=True, errors=\'coerce\')'
            elif is_timedelta_dtype(new_dtype):
                conversion_code = f'pd.to_timedelta({df_name}[\'{column_header}\'], errors=\'coerce\')'
        elif is_datetime_dtype(old_dtype):
            if is_bool_dtype(new_dtype):
                conversion_code = f'~{df_name}[\'{column_header}\'].isnull()'
            elif is_int_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'int\') / 10**9'
            elif is_float_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'int\').astype(\'float\') / 10**9'
            elif is_string_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].dt.strftime(\'%Y-%m-%d %X\')'
            elif is_datetime_dtype(new_dtype):
                pass
            elif is_timedelta_dtype(new_dtype):
                raise make_invalid_column_type_change_error(
                    column_header,
                    old_dtype,
                    new_dtype
                )
        elif is_timedelta_dtype(old_dtype):
            if is_bool_dtype(new_dtype):
                conversion_code = f'~{df_name}[\'{column_header}\'].isnull()'
            elif is_int_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].dt.total_seconds().astype(\'int\')'
            elif is_float_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].dt.total_seconds()'
            elif is_string_dtype(new_dtype):
                conversion_code = f'{df_name}[\'{column_header}\'].astype(\'str\')'
            elif is_datetime_dtype(new_dtype):
                raise make_invalid_column_type_change_error(
                    column_header,
                    old_dtype,
                    new_dtype
                )
            elif is_timedelta_dtype(new_dtype):
                pass

        return [f'{df_name}[\'{column_header}\'] = {conversion_code}']

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        column_header: str,
        old_dtype: str,
        new_dtype: str,
        df_names=None,
        **params
    ) -> str:
        return f'Changed {column_header} from {old_dtype} to {new_dtype}'


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Rename an existing saved analysis.
"""

from mitosheet.saved_analyses import rename_saved_analysis

RENAME_SAVED_ANALYSIS_UPDATE_EVENT = 'rename_analysis_update'
RENAME_SAVED_ANALYSIS_UPDATE_PARAMS = [
    'old_analysis_name',
    'new_analysis_name'
]

def execute_rename_saved_analysis_update(
        steps_manager,
        old_analysis_name,
        new_analysis_name
    ):
    """
    This function renames the saved analysis from the user's .mito folder
    """ 
    rename_saved_analysis(old_analysis_name, new_analysis_name)
    

RENAME_SAVED_ANALYSIS_UPDATE = {
    'event_type': RENAME_SAVED_ANALYSIS_UPDATE_EVENT,
    'params': RENAME_SAVED_ANALYSIS_UPDATE_PARAMS,
    'execute': execute_rename_saved_analysis_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains an analysis with the merge steps for tutorial v3
"""

MITO_TUTORIAL_CHECKPOINT_3 = {
    'name': 'checkpoint3',
    'saved_analysis': {
        "version": "0.1.96",
        "steps": {
            "1": {
                "step_version": 1,
                "step_type": "simple_import",
                "file_names": [
                    "AMTRAK-Stations-2010.csv",
                    "Zipcode-Data-2010.csv"
                ]
            },
            "2": {
                "step_version": 1,
                "step_type": "merge",
                "sheet_index_one": 0,
                "merge_key_one": "Zip",
                "selected_columns_one": [
                    "State",
                    "Checked_Bags",
                    "Zip"
                ],
                "sheet_index_two": 1,
                "merge_key_two": "Zip",
                "selected_columns_two": [
                    "Median_Income",
                    "Zip"
                ]
            }
        }
    }
}






#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Information about the frontend package of the widgets.
"""

from pathlib import Path
import os
import json
from mitosheet._version import __version__, get_package_json

module_name = get_package_json()['name']
module_version = '^' + __version__



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Decorators for sheet functions, that enable the sheet functions
to make sure the inputs they are getting are the inputs they want
"""
from typing import Union, Tuple
import pandas as pd
from functools import wraps, partial

from mitosheet.sheet_functions.types.utils import (
    get_mito_type,
    get_nan_indexes, put_nan_indexes_back
)
from mitosheet.errors import EditError, make_execution_error, make_invalid_arguments_error
from mitosheet.sheet_functions.types import SERIES_CONVERSION_FUNCTIONS


def handle_sheet_function_errors(sheet_function):
    """
    The first decorator that should be applied to every sheet function. Is 
    responsible for sensible error handling!
    """
    @wraps(sheet_function)
    def wrapped_f(*args):
        try:
            return sheet_function(*args)
        except EditError as e:
            raise e
        except:
            raise make_execution_error()
    
    return wrapped_f



def convert_arg_to_series_type(
        arg_index: int,
        arg_target_series_type, # Union[MITO_SERIES_TYPE, Literal['series']]
        on_uncastable_arg, # Literal['error', 'skip']
        on_uncastable_arg_element, #Union[Literal['error'], Tuple[Literal['default'], any]]
        optional=False
    ):
    """
    Wrapper for a sheet function that takes a fixed number of arguments, that takes
    - arg_index, which says which arg to force into which type
    - arg_target_series_type, or which series type to get all arguments into
    - on_uncastable_arg, which is what to do if an entire arg cannot be cast to this type
    - on_uncastable_arg_element, which is what to do when a single element in an arg
      cannot be cast. Include a default value with this function
    - optional, a boolean which allows the argument to be ignored if it is not 
      passed to the sheet function
    """
    def wrap(sheet_function):
        @wraps(sheet_function)
        def wrapped_sheet_function(*args):
            # Cast args to a list, so we can reassign to it (tuples are immutable)
            args = list(args)
            if optional and arg_index >= len(args):
                # If this is an optional argument, and is not passed, then we 
                # just ignore it
                pass
            elif arg_target_series_type == 'series':
                # If we just want any series, that's easy, we turn it into a series, if it is not 
                # already one
                mito_type = get_mito_type(args[arg_index])
                if not mito_type.endswith('series'):
                    args[arg_index] = pd.Series([args[arg_index]])
            else:
                # Otherwise, we actually convert
                conversion_function = SERIES_CONVERSION_FUNCTIONS[arg_target_series_type]
                try:
                    new_arg = conversion_function(args[arg_index], on_uncastable_arg_element=on_uncastable_arg_element)
                except:
                    if on_uncastable_arg_element == 'error':
                        raise make_invalid_arguments_error(sheet_function.__name__)
                if on_uncastable_arg == 'error' and None is new_arg:
                    raise make_invalid_arguments_error(sheet_function.__name__)
                args[arg_index] = new_arg
            return sheet_function(*args)  
        return wrapped_sheet_function
    return wrap


def convert_args_to_series_type(
        arg_target_series_type, #MITO_SERIES_TYPE,
        on_uncastable_arg, # Literal['error', 'skip'],
        on_uncastable_arg_element, # Union[Literal['error'], Tuple[Literal['default'], any]]
    ):
    """
    A decorator for functions like SUM or CONCAT, takes:
    - arg_target_series_type, or which series type to get all arguments into
    - on_uncastable_arg, which is what to do if an entire arg cannot be cast to this type
    - on_uncastable_arg_element, which is what to do when a single element in an arg
      cannot be cast. Include a default value with this function
    """

    def wrap(sheet_function):
        @wraps(sheet_function)
        def wrapped_sheet_function(*args):   
            conversion_function = SERIES_CONVERSION_FUNCTIONS[arg_target_series_type]
            # We partially apply the on_uncastable_arg_element, so that we can map the conversion function
            # over the arguments
            conversion_function = partial(conversion_function, on_uncastable_arg_element=on_uncastable_arg_element)
            new_args = map(conversion_function, args)

            # If any failed to cat, and on_uncastable_arg says to error, we error
            if on_uncastable_arg == 'error' and None in new_args:
                raise make_invalid_arguments_error(sheet_function.__name__)

            # Filter out the None values, as we don't want to send them to the function
            new_args = list(filter(lambda arg: arg is not None, new_args)) 
            return sheet_function(*new_args)        
        return wrapped_sheet_function
    return wrap


def filter_nans(sheet_function):
    """
    A decorator for functions that do not want to receive any NaN values
    in their input. For every passed series, filters out the indexes that
    is NaN in any series.
    """
    @wraps(sheet_function)
    def wrapped_f(*args):
        # before changing the types, determine the indexes that contain nan values
        nan_indexes = get_nan_indexes(*args)
            
        # for each arg that is an instance of a pd.series, filter out the nan_indexes
        new_args = [
            arg.loc[~nan_indexes].reset_index(drop=True) if isinstance(arg, pd.Series)
            else arg for arg in args
        ]
            
        result = sheet_function(*new_args)

        return put_nan_indexes_back(result, nan_indexes)
    
    return wrapped_f

def fill_nans(
        arg_index: int,
        new_value: any,
        optional=False
    ):
    """
    Wrapper for a sheet function that, for a specific argument, fills any NaN values with
    the passed new_value. 

    NOTE: this sometimes makes sense to call before the type casting decorators, and sometimes
    after (depending on what you are trying to accomplish). It does likely make sense to call 
    _after_ the type casting code, as 
    """
    def wrap(sheet_function):
        @wraps(sheet_function)
        def wrapped_sheet_function(*args):
            # Cast args to a list, so we can reassign to it (tuples are immutable)
            args = list(args)
            if optional and arg_index >= len(args):
                # If this is an optional argument, and is not passed, then we 
                # just ignore it
                pass

            arg = args[arg_index]
            
            # If it's a non-series, and it's a NaN, then just replace it
            if not isinstance(arg, pd.Series) and pd.isna(arg):
                args[arg_index] = new_value
            # Otherwise, if the argument is a series, then we fill all na values
            elif isinstance(arg, pd.Series):
                args[arg_index] = arg.fillna(new_value)
            
            return sheet_function(*args)  
        return wrapped_sheet_function
    return wrap

def cast_output(
        output_target_series_type # Union[MITO_SERIES_TYPE, Literal['first_input_type']]
    ):
    """
    Casts the output of the sheet function to the given type. If 'first_input_type' as given, then
    will cast the output to the type of the first passed argument. 

    If the cast to the output cannot occur or fails, then will return the output
    in the uncast type.

    NOTE: this decorator should always be the first decorator on a function, so that
    it can read in the input type! 
    """

    def wrap(sheet_function):
        @wraps(sheet_function)
        def wrapped_sheet_function(*args):
            if output_target_series_type == 'first_input_type':
                input_mito_type = get_mito_type(args[0])
                # Turn it into a series, if it is not
                if not input_mito_type.endswith('series'):
                    input_mito_type = get_mito_type(pd.Series([args[0]]))
                conversion_function = SERIES_CONVERSION_FUNCTIONS[input_mito_type]
            else:
                conversion_function = SERIES_CONVERSION_FUNCTIONS[output_target_series_type]
                        
            result = sheet_function(*args)
            try:
                return conversion_function(result, on_uncastable_arg_element='error')
            except:
                return result
            
        return wrapped_sheet_function
    return wrap


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
For going to a boolean series.
"""

from typing import Tuple, Union
import pandas as pd
import numpy as np

from mitosheet.sheet_functions.types.utils import BOOLEAN_SERIES, DATETIME_SERIES, NUMBER_SERIES, STRING_SERIES, get_mito_type


def to_boolean_series(
        unknown_object,
        on_uncastable_arg_element=('default', np.NaN), # Union[Literal['error'], Tuple[Literal['default'], any]]
    ):
    """
    Converts the given object to a boolean series. Note that on_uncastable_arg_element
    is irrelevant here, as anything can be turned into a boolean. 

    TODO: improve the string cast, it's dumb currently. It should detect true and True
    and false and False, and otherwise default to something reasonable?
    """
    from_type = get_mito_type(unknown_object)

    # If it is not a series, we put it in a series, and get the type again
    if not from_type.endswith('series'):
        unknown_object = pd.Series([unknown_object])
        from_type = get_mito_type(unknown_object)

    if from_type == BOOLEAN_SERIES:
        return unknown_object
    elif from_type == DATETIME_SERIES:
        # For now, we treat all dates as true, and NaN values as False
        return ~unknown_object.isna()
    elif from_type == NUMBER_SERIES:
        return unknown_object.astype('bool')
    elif from_type == STRING_SERIES:
        return unknown_object.astype('bool')
    else:
        return None


"""
conftest is the standard pytest testing config file, where
we store decorators we use
"""

from datetime import datetime
import json
import os
import pytest
from copy import deepcopy

from mitosheet import __version__
from mitosheet.user import is_running_test
from mitosheet.user.db import USER_JSON_PATH, get_user_field, get_user_json_object
from mitosheet.user.schemas import UJ_MITOSHEET_TELEMETRY, USER_JSON_DEFAULT
from mitosheet.user.schemas import (
    UJ_USER_JSON_VERSION, UJ_STATIC_USER_ID, UJ_USER_SALT, UJ_USER_EMAIL, 
    UJ_RECEIVED_TOURS, UJ_FEEDBACKS, UJ_MITOSHEET_CURRENT_VERSION, 
    UJ_MITOSHEET_LAST_UPGRADED_DATE, UJ_MITOSHEET_LAST_FIFTY_USAGES
)

today_str = datetime.today().strftime('%Y-%m-%d')


@pytest.fixture(scope="module", autouse=True)
def cleanup_files():
    """
    This fixture reads in the original user.json file that exists before these tests are
    run, deletes it, and then recreates it at the end. This allows us to test what happens 
    when the user.json file is in various states of out of date and disrepair.

    It also turns off the sending of logging to make sure we don't generate a huge number
    of new accounts on the frontend.

    It is notably autoused for this module, which means it runs before all users tests, and
    thus makes sure that running tests do not overwrite the current user.json you have.
    """
    with open(USER_JSON_PATH, 'r') as f:
        user_json = json.loads(f.read())

    os.remove(USER_JSON_PATH)

    yield # All tests in this user module run right here

    with open(USER_JSON_PATH, 'w+') as f:
        f.write(json.dumps(user_json))


def write_fake_user_json(user_json_object, **kwargs):
    """
    A helper function that can write a user_json_object,
    while also allowing keyword arguments to be passed
    for easy overwriting of default variables.
    """
    user_json_object = deepcopy(user_json_object)

    for field, value in kwargs.items():
        user_json_object[field] = value
    
    with open(USER_JSON_PATH, 'w+') as f:

        f.write(json.dumps(user_json_object))

def check_user_json(
        user_email='github@action.com',
        received_tours=[],
        feedbacks=[],
        mitosheet_current_version=__version__,
        mitosheet_last_upgraded_date=today_str,
        mitosheet_last_fifty_usages=[today_str],
    ):
    """
    This is the main helper function that does sanity checks about the 
    user_json object. 

    Pass keywords to check non-defaul values for the keys!
    """
    assert set(get_user_json_object().keys()) == set(USER_JSON_DEFAULT.keys())

    assert get_user_field(UJ_USER_JSON_VERSION) == 3
    assert len(get_user_field(UJ_STATIC_USER_ID)) > 0
    assert len(get_user_field(UJ_USER_SALT)) > 0
    assert (get_user_field(UJ_USER_EMAIL) == user_email or (get_user_field(UJ_USER_EMAIL) == 'github@action.com' and is_running_test()))
    assert get_user_field(UJ_RECEIVED_TOURS) == received_tours
    assert get_user_field(UJ_FEEDBACKS) == feedbacks
    assert get_user_field(UJ_MITOSHEET_CURRENT_VERSION) == mitosheet_current_version
    assert get_user_field(UJ_MITOSHEET_LAST_UPGRADED_DATE) == mitosheet_last_upgraded_date
    assert get_user_field(UJ_MITOSHEET_LAST_FIFTY_USAGES) == mitosheet_last_fifty_usages
    assert get_user_field(UJ_MITOSHEET_TELEMETRY) == True




from mitosheet.sheet_functions.types.utils import NUMBER_SERIES, get_mito_type
import pandas as pd
import json

def get_column_describe(event, steps_manager):
    """
    Sends back a string that can be parsed to a JSON object that
    contains _all_ the results from the series .describe function
    for the series at column_header in the df at sheet_index.
    """
    sheet_index = event['sheet_index']
    column_header = event['column_header']
    
    series: pd.Series = steps_manager.dfs[sheet_index][column_header]
    describe = series.describe()

    try:
        describe_obj = {}

        for index, row in describe.iteritems():
            # We turn all the items to strings, as some items are not valid JSON
            # e.g. some wacky numpy datatypes. This allows us to send all of this 
            # to the front-end.

            # If the series is a number, round the statistics so they look good.
            if get_mito_type(series) == NUMBER_SERIES:
                row = round(row, 2)

            describe_obj[index] = str(row)

        # We fill in some specific values that dont get filled by default
        describe_obj['count: NaN'] = str(series.isna().sum())

        # NOTE: be careful adding things here, as we dont want to destroy performance 
        if get_mito_type(series) == NUMBER_SERIES:
            describe_obj['median'] = str(round(series.median(), 2))
            describe_obj['sum'] = str(round(series.sum(), 2))

        return json.dumps(describe_obj)
    except:
        # As this is also a non-critical error, we don't want to display an error
        # message in the case of failure, so we just return nothing
        return ''


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains helpful functions for upgrading the user.json file
from previous versions to the current version.

An explination of how we used to do things:
1. We had a `user.json` file, which either didn't have a version number or was version 1
2. We added fields to the `user.json` file when the user loads it, 
   setting them equal to the new default values
3. We never made any backwards incompatible changes

However, we now want to make some backwards incompatible changes. We do that by having mulitple
schemas for the user.json file, and having an upgrade process that upgrades from one schema to 
the other.

1. For new users who show up, they will just get a default `user.json` file. 
   It will have the most up to date version number in it, obviously.

2. When old users show up, we compare the version of the user.json that they 
   have to the current version of the user.json.
   - If they have a user.json with _no version_ number or _version 1_, then we make sure they
     have the "final schema" for version 1. We only need to do this because of how we used to 
     change version 1 without making new versions.
   - Then, if they have an old version of the user.json, then we actually do an upgrade from one 
     version to the next. Simply using an upgrade script that converts.

_Any change the user.json format must be met with a version bump._
"""
from mitosheet.user.schemas import (
    UJ_INTENDED_BEHAVIOR, UJ_CLOSED_FEEDBACK, UJ_MITOSHEET_LAST_FIVE_USAGES, UJ_MITOSHEET_TELEMETRY,
    UJ_USER_JSON_VERSION, UJ_MITOSHEET_LAST_FIFTY_USAGES
)
from mitosheet.user.schemas import USER_JSON_DEFAULT, USER_JSON_VERSION_1
from mitosheet.user.db import get_user_json_object, set_user_json_object

def try_upgrade_to_final_user_json_version_1(user_json_object):
    """
    With user_json_version = 1, we used to make changes to the user.json 
    without changing the version number. We simply just added a bunch
    of new fields.

    Thus, if a user is on user_json_version = 1, before upgrading the user 
    json to a later version, we first make sure they have all the fields
    that the final user_json_version = 1 had.

    NOTE: this also handles the user.json that is written by the installer, 
    which is effectively just the static_user_id and nothing else.
    """

    # If the user is on an incredibly old version, we might not have a version number
    # in which case we just set it equal to 1, and continue with the upgrade
    if UJ_USER_JSON_VERSION not in user_json_object:
        user_json_object[UJ_USER_JSON_VERSION] = 1
    
    # We only need to upgrade to the final version of 1 in the case
    # that they are actually in version 1!
    if user_json_object[UJ_USER_JSON_VERSION] == 1:
            
        # Then we just make sure that the user.json has all the fields it needs defined
        # in the final version one schema, and if they are not defined, it sets them 
        # to the default values provided
        for field, default_value in USER_JSON_VERSION_1.items():
            if field not in user_json_object:
                user_json_object[field] = default_value

    return user_json_object

def upgrade_final_user_json_version_1_to_2(final_user_json_version_1):
    """
    Two changes in this version: 
    1. removes the intended_behavior field, as we no longer user it
    2. removes the closed_feedback field, as we no longer user it
    3. renames mitosheet_last_five_usages -> mitosheet_last_fifty_usages, 
       so we can now track 50 usages
    """
    if final_user_json_version_1[UJ_USER_JSON_VERSION] != 1:
        return final_user_json_version_1

    # 0: bump the version number
    final_user_json_version_1[UJ_USER_JSON_VERSION] = 2
    
    # 1: Remove intended behavior field
    del final_user_json_version_1[UJ_INTENDED_BEHAVIOR]

    # 2: Remove closed_feedback field
    del final_user_json_version_1[UJ_CLOSED_FEEDBACK]

    # 3: Rename mitosheet_last_five_usages -> mitosheet_last_fifty_usages
    final_user_json_version_1[UJ_MITOSHEET_LAST_FIFTY_USAGES] = final_user_json_version_1[UJ_MITOSHEET_LAST_FIVE_USAGES]
    del final_user_json_version_1[UJ_MITOSHEET_LAST_FIVE_USAGES]

    return final_user_json_version_1

def upgrade_user_json_version_2_to_3(user_json_version_2):
    """
    Just adds the user json telemtry field
    """
    # First, bump the version number
    user_json_version_2[UJ_USER_JSON_VERSION] = 3

    # Then, set new field
    user_json_version_2[UJ_MITOSHEET_TELEMETRY] = True
    return user_json_version_2

def try_upgrade_user_json_to_current_version():
    user_json_object = get_user_json_object()

    # If we don't need to upgrade, don't upgrade
    if UJ_USER_JSON_VERSION in user_json_object and user_json_object[UJ_USER_JSON_VERSION] == USER_JSON_DEFAULT[UJ_USER_JSON_VERSION]:
        return

    # Try to upgrade it to the final version of 1, if it is not already there
    try_upgrade_to_final_user_json_version_1(user_json_object)

    # Then, upgrade from version 1 to version 2, which is the 
    # most up-to-date version
    if user_json_object[UJ_USER_JSON_VERSION] == 1:
        user_json_object = upgrade_final_user_json_version_1_to_2(user_json_object)
    if user_json_object[UJ_USER_JSON_VERSION] == 2:
        user_json_object = upgrade_user_json_version_2_to_3(user_json_object)

    set_user_json_object(user_json_object)





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
After reading in the arguments passed to the frontend,
this update figures out which of them were dataframes
and which of them were file paths, and updates the 
df names in the steps properly.
"""

from mitosheet.step_performers.import_steps.simple_import import get_dataframe_names

ARGS_UPDATE_EVENT = 'args_update'
ARGS_UPDATE_PARAMS = ['args']


def is_str_df_name(arg):
    """
    Helper function that returns true if this argument
    name is in fact a file path - aka it is passed
    as a string to the mitosheet (not a dataframe
    variable directly).
    """
    return arg.startswith('\'') or arg.startswith('\"')

def execute_args_update(
        steps_manager,
        args
    ):
    """
    Changes the df_names in the current step - from information
    from the frontend.

    The information from the frontend is a list of all the parameters
    to the mitosheet.sheet call, which may contain both dataframe names
    as well as as strings. 

    For the dataframe names, we take them as is. For the strings, we 
    strip the quotes, and turn them into valid df names. NOTE: the df
    names we turn it into _must be the same as the names transpiled_
    by the preprocess read file paths preprocessing step!
    """
    # We send / don't do anything with empty df_name updates
    # as we want the sheet to refresh in this case with no errors
    if len(args) > 0:
        # Get all the string arguments, and strip off the quotes
        str_args = [arg for arg in args if is_str_df_name(arg)]
        str_args = [str_arg[1:-1] for str_arg in str_args]

        # Then, we turn these into dataframe names.
        # NOTE: there is potentially a bug if a user passes in a dataframe
        # with the same name as the the result of the return of this function,
        # but we ignore this for now. E.g. mitosheet.sheet('df1.csv', 'df1_csv') 
        # will cause only variable to exist.
        str_df_names = get_dataframe_names(str_args, [])

        final_names = []
        used_str_df_names = 0
        for arg in args:
            if is_str_df_name(arg):
                final_names.append(str_df_names[used_str_df_names])
                used_str_df_names += 1
            else:
                final_names.append(arg)

        # Finially, we don't add more names than there are dataframes (as this is clearly
        # nonsense), and thus this allows us to filter out Nones that are passed at the 
        # end of the arguments (not creating phantom tabs that cannot be clicked)
        steps_manager.curr_step.post_state.df_names = final_names[:len(steps_manager.curr_step.dfs)]


ARGS_UPDATE = {
    'event_type': ARGS_UPDATE_EVENT,
    'params': ARGS_UPDATE_PARAMS,
    'execute': execute_args_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import deepcopy
from mitosheet.state import State
from typing import Any, Dict, List, Optional
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_column_delete_error
)

class DeleteColumnStepPerformer(StepPerformer):
    """"
    A delete_column step, which allows you to delete a column
    from a dataframe.
    """
    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'delete_column'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Deleted a Column'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'delete_column_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        column_header: str,
        **params
    ) -> State:
        # if the sheet doesn't exist, throw an error
        if not prev_state.does_sheet_index_exist_within_state(sheet_index):
            raise make_no_sheet_error(sheet_index)

        # Error if the column does not exist
        if column_header not in prev_state.column_metatype[sheet_index]:
            raise make_no_column_error([column_header])
        
        # Error if there are any columns that currently rely on this column
        if len(prev_state.column_evaluation_graph[sheet_index][column_header]) > 0:
            raise make_invalid_column_delete_error(
                column_header,
                list(prev_state.column_evaluation_graph[sheet_index][column_header])
            )

        # Make a post state, that is a deep copy
        post_state = deepcopy(prev_state)
        
        # Actually drop the column
        df = post_state.dfs[sheet_index]
        df.drop(column_header, axis=1, inplace=True)

        # And then update all the state variables removing this column from the state
        del post_state.column_metatype[sheet_index][column_header]
        del post_state.column_type[sheet_index][column_header]
        del post_state.column_spreadsheet_code[sheet_index][column_header]
        del post_state.column_python_code[sheet_index][column_header]
        del post_state.column_evaluation_graph[sheet_index][column_header]
        # We also have to delete the places in the graph where this node is 
        for dependents in post_state.column_evaluation_graph[sheet_index].values():
            if column_header in dependents:
                dependents.remove(column_header)
        
        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        column_header: str
    ) -> List[str]:
        df_name = post_state.df_names[sheet_index]
        return [f'{df_name}.drop(\'{column_header}\', axis=1, inplace=True)']

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        column_header: str,
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            df_name = df_names[sheet_index]
            return f'Deleted column {column_header} from {df_name}'
        return f'Deleted column {column_header}'


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

from copy import deepcopy
from mitosheet.step_performers.step_performer import StepPerformer
from typing import Any, Dict, List, Optional
from mitosheet.state import State
from mitosheet.sheet_functions.types.utils import get_mito_type
from mitosheet.topological_sort import creates_circularity, subgraph_from_starting_column_header, topological_sort_columns
from mitosheet.sheet_functions import FUNCTIONS
from mitosheet.parser import parse_formula

from mitosheet.errors import (
    make_circular_reference_error, 
    make_no_column_error, 
    make_operator_type_error, 
    make_unsupported_function_error, 
    make_wrong_column_metatype_error
)

class SetColumnFormulaStepPerformer(StepPerformer):
    """
    A set_column_formula step, which allows you to set the formula
    of a given column in the sheet (and then recalculates this column)
    and it's dependents.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'set_column_formula'

    @classmethod
    def step_display_name(cls) -> str:
        return 'Set Column Formula'
    
    @classmethod
    def step_event_type(cls) -> str:
        return 'set_column_formula_edit'

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        sheet_index = params['sheet_index']
        column_header = params['column_header']
        params['old_formula'] = prev_state.column_spreadsheet_code[sheet_index][column_header]
        
        # If the user submits an empty formula, we just set it equal to zero
        if params['new_formula'] == '':
            params['new_formula'] = '=0'
        else:
            try:
                # Try and parse the formula, letting it throw errors if it
                # is invalid
                parse_formula(params['new_formula'], column_header, throw_errors=True)
            except:
                params['new_formula'] = _get_fixed_invalid_formula(params['new_formula'], column_header)

        # By default, we don't do anything with the saturate
        return params

    @classmethod
    def execute(
        cls,
        prev_state: State,
        sheet_index: int,
        column_header: str,
        old_formula: str,
        new_formula: str,
        **params
    ) -> State:
        if column_header not in prev_state.column_metatype[sheet_index]:
            raise make_no_column_error([column_header])

        # First, we check the column_metatype, and make sure it's a formula
        if prev_state.column_metatype[sheet_index][column_header] != 'formula':
            raise make_wrong_column_metatype_error(column_header)

        # If nothings changed, there's no work to do
        if (old_formula == new_formula):
            return

        # Then we try and parse the formula
        new_python_code, new_functions, new_dependencies = parse_formula(
            new_formula, 
            column_header
        )
        
        # We check that the formula doesn't reference any columns that don't exist
        missing_columns = new_dependencies.difference(prev_state.column_metatype[sheet_index].keys())
        if any(missing_columns):
            raise make_no_column_error(missing_columns)

        # The formula can only reference known formulas
        missing_functions = new_functions.difference(set(FUNCTIONS.keys()))
        if any(missing_functions):
            raise make_unsupported_function_error(missing_functions)

        # Then, we get the list of old column dependencies and new dependencies
        # so that we can update the graph
        old_python_code, old_functions, old_dependencies = parse_formula(old_formula, column_header)

        # Before changing any variables, we make sure this edit didn't
        # introduct any circularity
        circularity = creates_circularity(
            prev_state.column_evaluation_graph[sheet_index], 
            column_header,
            old_dependencies,
            new_dependencies
        )
        if circularity:
            raise make_circular_reference_error()

        # We check out a new step
        post_state = deepcopy(prev_state)

        # Update the column formula, and then execute the new formula graph
        _update_column_formula_in_step(post_state, sheet_index, column_header, old_formula, new_formula)
        _execute(post_state, post_state.dfs[sheet_index], sheet_index)

        # Finially, update the type of the filters of this column, for all the filters
        # TODO: fix bug where we have to update downstream types, but note that
        # it would just be really nice if we didn't have to store the type here, and could
        # just get it dynamically...
        new_type = get_mito_type(post_state.dfs[sheet_index][column_header])
        post_state.column_type[sheet_index][column_header] = new_type
        post_state.column_filters[sheet_index][column_header]['filters'] = [
            {'type': new_type, 'condition': filter_['condition'], 'value': filter_['value']} 
            for filter_ in prev_state.column_filters[sheet_index][column_header]['filters']
        ]

        return post_state, None

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        execution_data: Optional[Dict[str, Any]],
        sheet_index: int,
        column_header: str,
        old_formula: str,
        new_formula: str
    ) -> List[str]:
        """
        Transpiles an set_column_formula step to python code!
        """
        code = []

        # We only look at the sheet that was changed, and sort the columns, taking only
        # those downstream from the changed columns
        subgraph = subgraph_from_starting_column_header(post_state.column_evaluation_graph[sheet_index], column_header)
        topological_sort = topological_sort_columns(subgraph)

        # We compile all of their formulas
        for column in topological_sort:
            column_formula_changes = post_state.column_python_code[sheet_index][column]
            if column_formula_changes != '':
                # We replace the data frame in the code with it's parameter name!
                # NOTE: we check for df[ to increase the odds that we don't replace
                # something other than the dataframe name itself (e.g. replacing a column
                # name with the letters "df" inside of them
                column_formula_changes = column_formula_changes.strip().replace('df[', f'{post_state.df_names[sheet_index]}[')
                code.append(column_formula_changes)

        return code

    @classmethod
    def describe(
        cls,
        sheet_index: int,
        column_header: str,
        old_formula: str,
        new_formula: str,
        df_names=None,
        **params
    ) -> str:
        if df_names is not None:
            df_name = df_names[sheet_index]
            return f'Set {column_header} in {df_name} to {new_formula}'
        return f'Set {column_header} to {new_formula}'


def _update_column_formula_in_step(
        post_state: State,
        sheet_index: int,
        column_header: str,
        old_formula: str,
        new_formula: str
    ):
    """
    A  helper function for updating the formula of a column. It assumes
    that the passed information is all _correct_ and will not:
    1. Introduce a circular reference error.
    2. Add an invalid formula.

    It DOES NOT "reexecute" the dataframes, just updates the state variables.
    """

    new_python_code, _, new_dependencies = parse_formula(
        new_formula, 
        column_header
    )

    _, _, old_dependencies = parse_formula(
        old_formula, 
        column_header
    )

    post_state.column_spreadsheet_code[sheet_index][column_header] = new_formula
    post_state.column_python_code[sheet_index][column_header] = new_python_code

    # Update the column dependency graph
    for old_dependency in old_dependencies:
        post_state.column_evaluation_graph[sheet_index][old_dependency].remove(column_header)
    for new_dependency in new_dependencies:
        post_state.column_evaluation_graph[sheet_index][new_dependency].add(column_header)


def _get_fixed_invalid_formula(new_formula, address):
    """
    A helper function that, given a formula, will try and fix
    any common errors with it. 

    Currently, the common errors are just checking if the formula
    is missing some number of parenthases at the end.

    Returns the fixed formula. If it cannot fix the formula, then 
    returns the original formula (so that execution continues as 
    normal, to report the error to the user).
    """
    POTENTIAL_VALID_FORMULAS = [
        # Check the original formula, just in case it actually is valid
        new_formula,
        new_formula + ')',
        new_formula + '))',
        new_formula + ')))',
    ]
    for fixed_formula in POTENTIAL_VALID_FORMULAS:
        try:
            # Parse the formula, and return if it is valid
            parse_formula(fixed_formula, address, throw_errors=True)
            return fixed_formula
        except:
            pass
    
    return new_formula


def get_details_from_operator_type_error(error):
    """
    We detect operator errors by checking the error string, which has the format:

    If you write 1 + 'true'
    - unsupported operand type(s) for /: 'int' and 'str' 

    If you write 1 + pd.Series(['true'])
    - can only concatenate str (not "int") to str

    If you write 10 ^ pd.Series(['true'])
    - Cannot perform 'rxor' with a dtyped [object] array and scalar of type [bool]
    If you write pd.Series(['true']) ^ 10
    - Cannot perform 'xor' with a dtyped [object] array and scalar of type [bool]

    ^ NOTE: the above two errors makes pretty much no sense at all, but it appears
    to be casting the number to a boolean? So we choose to handle this as a number,
    as this is when you would be 

    Returns None if the passed error is not an operator type error. Otherwise, returns
    a triple of (operator, first argument type, second argument type).
    """
    error_message = str(error)
    # Handle case where standard python primitive types are used incorrectly
    if error_message.startswith('unsupported operand type'):
        # Then, we get the specific operator by checking right before the :
        operator = error_message.split(':')[0].split(' ')[-1].strip()
        # And we get the specific types - see above format for how this parsing works!
        arg_one_type = error_message.split(':')[-1].split('\'')[1]
        arg_two_type = error_message.split(':')[-1].split('\'')[-2]
        return (operator, arg_one_type, arg_two_type)
    # Handle case where pandas dataframes are part of the operator, for a concatenate
    if error_message.startswith('can only concatenate str'):
        # We cannot tell the ordered here (e.g. we don't know if a string was concated
        # to an integer, or the other way around) b/c the error doesn't say so, so we just
        # do our best 
        other_type = error_message.split('"')[1]
        return ('+', 'str', other_type)
    # Handle a ^
    if error_message.startswith('Cannot perform \'rxor\' with a dtyped') or error_message.startswith('Cannot perform \'xor\' with a dtyped'):
        # This error message might be totally wrong, but there is some weird
        # casting that goes on in the error message that makes it hard to tell. As such,
        # we report the most likely error.
        return ('^', 'number', 'str')


    return None


def _execute(post_state, df, sheet_index):
    """
    Executes the given state variables for  
    """

    topological_sort = topological_sort_columns(post_state.column_evaluation_graph[sheet_index])

    for column_header in topological_sort:
        # Exec the code, where the df is the original dataframe
        # See explination here: https://www.tutorialspoint.com/exec-in-python
        try:
            exec(
                post_state.column_python_code[sheet_index][column_header],
                {'df': df}, 
                FUNCTIONS
            )
        except TypeError as e:
            # We catch TypeErrors specificially, so that we can case on operator errors, to 
            # give better error messages
            operator_type_error_details = get_details_from_operator_type_error(e)
            if operator_type_error_details is not None:
                # If there is an operator error, we handle it specially, to give the user
                # more information about how to recover
                raise make_operator_type_error(*operator_type_error_details)
            else:
                # If it's not an operator error, we just propagate the error up
                raise e


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains an analysis with the import steps for tutorial v3
"""

MITO_TUTORIAL_CHECKPOINT_2 = {
    'name': 'checkpoint2',
    'saved_analysis': {
        "version": "0.1.96", 
        "steps": {
            "1": {
                "step_version": 1, 
                "step_type": "simple_import", 
                "file_names": ["AMTRAK-Stations-2010.csv", "Zipcode-Data-2010.csv"],
            }
        }
    }
}



from abc import ABC, abstractmethod
from mitosheet.state import State
from typing import Dict, List
 
class StepPerformer(ABC, object):
    """
    The abstract base class for a step generator, which are the set
    of classes that execute, transpile, and describe steps. 
    """

    @classmethod
    @abstractmethod
    def step_version(cls) -> int:
        """
        Returns the version of the step. Changes when the parameters
        of the step change.
        """
        pass

    @classmethod
    @abstractmethod
    def step_type(cls) -> str:
        """
        The name of the step used internally. If you change this, you must upgrade
        the step and bump the version.
        """
        pass

    @classmethod
    @abstractmethod
    def step_display_name(cls) -> str:
        """
        The name of the step that is displayed to the user
        in the frontend.
        """
        pass
    
    @classmethod
    @abstractmethod
    def step_event_type(cls) -> str:
        """
        The type of the edit event that generates this type of step.
        """
        pass

    @classmethod
    def saturate(cls, prev_state: State, params) -> Dict[str, str]:
        """
        Given the parameters of the step, will saturate the event with
        more parameters based on the passed prev_state. 
        """
        # By default, we don't do anything with the saturate
        return params

    @classmethod
    @abstractmethod
    def execute(cls, prev_state: State, **params) -> State:
        """
        Execute always returns the post_state, and optionally returns a dictionary
        of execution_data, which is data that may be useful to the transpiler in
        transpiling the code.
        """
        pass

    @classmethod
    @abstractmethod
    def transpile(cls, prev_state: State, post_state: State, **params) -> List[str]:
        """
        Returns a list of the Python code lines that corresponds to this 
        step being executed.
        """
        pass

    @classmethod
    @abstractmethod
    def describe(cls, **params) -> str:
        """
        Describes the step in a single sentence, usually for comments or for presentation
        to the user in the frontend.
        """
        pass


"""
Utilities for creating and initializing the user.json
file with the current schema
"""

from datetime import datetime
from mitosheet.user.utils import is_local_deployment
from mitosheet.utils import get_random_id
import os
import json

from mitosheet._version import __version__
from mitosheet.user.schemas import (
    UJ_STATIC_USER_ID, UJ_USER_EMAIL, UJ_MITOSHEET_CURRENT_VERSION, 
    UJ_MITOSHEET_LAST_UPGRADED_DATE, UJ_MITOSHEET_LAST_FIFTY_USAGES
)
from mitosheet.user.schemas import GITHUB_ACTION_EMAIL, GITHUB_ACTION_ID, USER_JSON_DEFAULT
from mitosheet.user.upgrade import try_upgrade_user_json_to_current_version
from mitosheet.user.db import MITO_FOLDER, USER_JSON_PATH, get_user_json_object, set_user_field, get_user_field

def is_user_json_exists_and_valid_json():
    """
    Helper function that determines if the current user json both
    exists and is valid json
    """
    if not os.path.exists(USER_JSON_PATH):
        return False

    try:
        with open(USER_JSON_PATH, 'r') as f:
            json.loads(f.read())
        return True
    except:
        return False


def try_create_user_json_file():
    # Create the mito folder if it does not exist
    if not os.path.exists(MITO_FOLDER):
        os.mkdir(MITO_FOLDER)

    # We create a user.json file if it does not exist, or if it
    # is invalid (e.g. it is not parseable JSON).
    if not is_user_json_exists_and_valid_json():
        # First, we write an empty default object
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(USER_JSON_DEFAULT))

        # Then, we take special care to put all the testing/CI enviornments 
        # (e.g. Github actions) under one ID and email
        from mitosheet.user.utils import is_running_test
        if is_running_test():
            set_user_field(UJ_STATIC_USER_ID, GITHUB_ACTION_ID)
            set_user_field(UJ_USER_EMAIL, GITHUB_ACTION_EMAIL)


def initialize_user(identify=True):
    """
    Internal helper function that gets called every time mitosheet 
    is imported.

    It:
    1. Creates the user.json if it does not exist (though it usually does, from the installer)
    2. Upgrades the user.json to the most up to date format
    3. Updates the user.json file with any usage or upgrading, logging anything interesting.
    
    By default, also identifies the user, but may not do this in some cases (when identify
    if False), as the user may just be turning logging off or something.
    """
    # Try to create the user.json file, if it does not already exist
    try_create_user_json_file()

    # Try to upgrade it to the most up to date format, if it is not already there
    try_upgrade_user_json_to_current_version()

    # Then, we check if Mito has been upgraded since it was last imported
    # and if it has been upgraded, we upgrade the version and the upgrade date
    # NOTE: we only log this if the user is local, as users who are not local
    # get upgraded on our infrastructure anyways
    mitosheet_current_version = get_user_field(UJ_MITOSHEET_CURRENT_VERSION)
    if mitosheet_current_version != __version__ and is_local_deployment():
        from mitosheet.mito_analytics import log
        set_user_field(UJ_MITOSHEET_CURRENT_VERSION, __version__)
        set_user_field(UJ_MITOSHEET_LAST_UPGRADED_DATE, datetime.today().strftime('%Y-%m-%d'))
        # Log the upgrade. Note that this runs when the user _actually_ changes
        # the version of mitosheet that they are using, not just when they 
        # click the upgrade button in the app (although clicking this upgrade
        # button will stop the upgrade popup from showing up)
        log('upgraded_mitosheet', {'old_version': mitosheet_current_version, 'new_version': __version__})

    # We also note this import as a Mito usage, making sure to only 
    # mark this as usage once per day
    last_fifty_usages = get_user_field(UJ_MITOSHEET_LAST_FIFTY_USAGES)
    if len(last_fifty_usages) == 0 or last_fifty_usages[-1] != datetime.today().strftime('%Y-%m-%d'):
        last_fifty_usages.append(datetime.today().strftime('%Y-%m-%d'))
    # Then, we take the 50 most recent (or as many as there are), and save them
    if len(last_fifty_usages) < 50:
        most_recent_fifty = last_fifty_usages
    else: 
        most_recent_fifty = last_fifty_usages[-50:]
    set_user_field(UJ_MITOSHEET_LAST_FIFTY_USAGES, most_recent_fifty)

    # Reidentify the user, just in case things have changed
    # but only if we were told to identify
    if identify:
        from mitosheet.mito_analytics import identify
        identify()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""
The mitosheet package is distributed under the mitosheet and mitosheet3
package names on pip. The package.json will tell you all you need to 
know about which one we are in currently using.

As such, this setup.py script reads in the package.json and sets up
the proper package.
"""

from __future__ import print_function
from glob import glob
from os.path import join as pjoin
import json
import setuptools
from pathlib import Path


from jupyter_packaging import (
    create_cmdclass,
    install_npm,
    ensure_targets,
    combine_commands,
    skip_if_exists,
    ensure_python,
)

from setuptools import setup

HERE = Path(__file__).parent.resolve()

package_json = json.loads(open('package.json').read())
lab_path = Path(pjoin(HERE, 'mitosheet', 'labextension'))


# The name of the project
name = package_json['name']

# Ensure a valid python version
ensure_python('>=3.4')


if name == 'mitosheet':
    # Get our version, which we just read 
    version = package_json['version']

    # Representative files that should exist after a successful build
    jstargets = [
        pjoin(HERE, 'lib', 'plugin.js'),
    ]

    package_data_spec = {
        name: [
            'labextension/*.tgz'
        ]
    }

    data_files_spec = [
        ('share/jupyter/lab/extensions', lab_path, '*.tgz'),
    ]


    cmdclass = create_cmdclass('jsdeps', package_data_spec=package_data_spec,
        data_files_spec=data_files_spec)
    cmdclass['jsdeps'] = combine_commands(
        install_npm(HERE, build_cmd='build:all'),
        ensure_targets(jstargets),
    )

    setup_args = dict(
        name            = name,
        description     = 'The Mito Spreadsheet',
        version         = version,
        scripts         = glob(pjoin('scripts', '*')),
        cmdclass        = cmdclass,
        packages        = setuptools.find_packages(exclude=['deployment']),
        author          = 'Mito',
        author_email    = 'naterush1997@gmail.com',
        license         = package_json["license"],
        url             = 'https://github.com/mito/mito',
        platforms       = "Linux, Mac OS X, Windows",
        keywords        = ['Jupyter', 'Widgets', 'IPython'],
        classifiers     = [
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: ' + package_json["license"],
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Framework :: Jupyter',
        ],
        include_package_data = True,
        # NOTE: this should be the same as the INSTALL_REQUIRES variable in
        # the mitoinstaller package
        install_requires = [
            # We require jupyterlab 2.0
            'jupyterlab>=2.0,<3.0,!=2.3.0,!=2.3.1', # there are css incompatabilities on version 2.3.1 and 2.3.0
            'ipywidgets>=7.0.0',
            'pandas>=1.1.0',
            # We don't need to lock an analytics-python version, as this library
            # is stabilized and mature
            'analytics-python',
            # Graphing libraries
            'plotly==4.14.3'
        ],
        extras_require = {
            'test': [
                'pytest',
                'flake8',
            ],
            'deploy': [
                'wheel', 
                'twine',
                "jupyter_packaging"
            ]
        },
        entry_points = {
        },
        long_description="""
            To learn more about Mito, checkout out our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
            Before installing Mito \n\n
            1. Check that you have Python 3.6 or above. To check your version of Python, open a new terminal, and type python3 --version. If you need to install or update Python, restart your terminal after doing so.\n\n
            2. Check that you have Node installed.To check this, open a new terminal, and type node -v.  It should print a version number. If you need to install Node, restart your terminal after doing so.\n\n
            3. Mito works in Jupyter Lab 2.0 only. We do not yet support Google Collab, VSCode, or Jupyter Lab 3.0.\n\n
            4. Checkout our terms of service and privacy policy. By installing Mito, you're agreeing to both of them. Please contact us at aarondr77 (@) gmail.com with any questions.\n\n
            Installation Instructions \n\n
            For more detailed installation instructions, see our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
            1. pip install mitoinstaller\n\n
            2. python -m mitoinstaller install
            """,
        long_description_content_type='text/markdown'
    )
elif name == 'mitosheet3':
    # Representative files that should exist after a successful build
    jstargets = [
        str(lab_path / "package.json"),
    ]

    package_data_spec = {
        'mitosheet': ["*"],
    }

    labext_name = "mitosheet3"

    data_files_spec = [
        ("share/jupyter/labextensions/%s" % labext_name, str(lab_path), "**"),
        ("share/jupyter/labextensions/%s" % labext_name, str(HERE), "install.json"),
    ]

    cmdclass = create_cmdclass("jsdeps",
        package_data_spec=package_data_spec,
        data_files_spec=data_files_spec
    )

    js_command = combine_commands(
        install_npm(HERE, build_cmd="build:prod", npm=["jlpm"]),
        ensure_targets(jstargets),
    )

    is_repo = (HERE / ".git").exists()
    if is_repo:
        cmdclass["jsdeps"] = js_command
    else:
        cmdclass["jsdeps"] = skip_if_exists(jstargets, js_command)

    setup_args = dict(
        name                    = name,
        version                 = package_json["version"],
        url                     = package_json["homepage"],
        author                  = package_json["author"]["name"],
        author_email            = package_json["author"]["email"],
        description             = package_json["description"],
        license                 = package_json["license"],
        long_description="""
            To learn more about Mito, checkout out our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
            Before installing Mito \n\n
            1. Check that you have Python 3.6 or above. To check your version of Python, open a new terminal, and type python3 --version. If you need to install or update Python, restart your terminal after doing so.\n\n
            2. Checkout our terms of service and privacy policy. By installing Mito, you're agreeing to both of them. Please contact us at aarondr77 (@) gmail.com with any questions.\n\n
            Installation Instructions \n\n
            For more detailed installation instructions, see our documentation: https://docs.trymito.io/getting-started/installing-mito\n\n
            1. pip install mitoinstaller\n\n
            2. python -m mitoinstaller install
        """,
        long_description_content_type = "text/markdown",
        cmdclass                 = cmdclass,
        packages                 = setuptools.find_packages(exclude=['deployment']),
        install_requires=[        
            "jupyterlab~=3.0",
            'ipywidgets>=7.0.0',
            # In JLab 3, we move to needing to install the jupyterlab-widgets package, which
            # is equivlaent to the @jupyter-widgets/jupyterlab-manager extension for jlab 2
            'jupyterlab-widgets==1.0.0',
            'pandas>=1.1.0',
            'analytics-python',
            # Graphing libraries
            'plotly==4.14.3'
        ],
        extras_require = {
            'test': [
                'pytest',
                'flake8',
            ],
            'deploy': [
                'wheel', 
                'twine',
                "jupyter_packaging"
            ]
        },
        zip_safe                = False,
        include_package_data    = True,
        python_requires         = ">=3.6",
        platforms               = "Linux, Mac OS X, Windows",
        keywords                = ["Jupyter", "JupyterLab", "JupyterLab3"],
        classifiers             = [
            "License :: " + package_json["license"],
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Framework :: Jupyter",
        ],
    )

if __name__ == '__main__':
    setup(**setup_args)



"""
The main entry point for the mitosheet package, this command
line interface allows you to set some toggles in the user.json
"""
from mitosheet.user.db import set_user_field
import sys
from mitosheet.user import initialize_user
from mitosheet.user.schemas import UJ_MITOSHEET_TELEMETRY, UJ_USER_EMAIL

def main():
    """
    Currently, the main usage of this function is:
    python -m mitosheet turnofflogging
    python -m mitosheet turnonlogging

    We also have 
    python -m mitosheet clearemail
    """
    # Make sure the user is initalized first, but do not identify
    # then, in case they are turning off logging
    initialize_user(identify=False)

    # Then, if we are being told to turn off logging, turn off logging
    if len(sys.argv) > 1:
        if sys.argv[-1] == 'turnofflogging':
            print("Turning off all logging")
            set_user_field(UJ_MITOSHEET_TELEMETRY, False)
            print("Logging turned off!")
        if sys.argv[-1] == 'turnonlogging':
            print("Turning on all logging")
            set_user_field(UJ_MITOSHEET_TELEMETRY, True)
            print("Logging turned on!")
        if sys.argv[-1] == 'clearemail':
            print("Clearing email")
            set_user_field(UJ_USER_EMAIL, '')
            print("Email cleared")


if __name__ == '__main__':
    main()


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.

"""
Main file containing the mito widget.
"""
from mitosheet.user.schemas import UJ_FEEDBACKS, UJ_MITOSHEET_LAST_FIFTY_USAGES
from mitosheet.user.db import get_user_field
from mitosheet.utils import get_new_id
from mitosheet.api import API
from mitosheet.mito_analytics import log, log_event_processed, log_recent_error
import pandas as pd

from ipywidgets import DOMWidget
import traitlets as t

from mitosheet._frontend import module_name, module_version
from mitosheet.errors import EditError, get_recent_traceback_as_list
from mitosheet.saved_analyses import (
    read_and_upgrade_analysis, 
    write_analysis,
    saved_analysis_names_json
)

from mitosheet.steps_manager import StepsManager

from mitosheet.user import is_local_deployment, should_upgrade_mitosheet
from mitosheet.data_in_mito import DataTypeInMito


class MitoWidget(DOMWidget):
    """
        The MitoWidget holds all of the backend state for the Mito extension, and syncs
        the state with the frontend widget. 
    """
    _model_name = t.Unicode('ExampleModel').tag(sync=True)
    _model_module = t.Unicode(module_name).tag(sync=True)
    _model_module_version = t.Unicode(module_version).tag(sync=True)
    _view_name = t.Unicode('ExampleView').tag(sync=True)
    _view_module = t.Unicode(module_name).tag(sync=True)
    _view_module_version = t.Unicode(module_version).tag(sync=True)

    is_local_deployment = t.Bool(True).tag(sync=True)
    analysis_name = t.Unicode('').tag(sync=True)
    curr_step_idx = t.Int(0).tag(sync=True)
    sheet_json = t.Unicode('').tag(sync=True)
    code_json = t.Unicode('').tag(sync=True)
    df_names_json = t.Unicode('').tag(sync=True)
    df_sources_json = t.Unicode('').tag(sync=True)
    df_shape_json = t.Unicode('').tag(sync=True)
    saved_analysis_names_json = t.Unicode('').tag(sync=True)
    column_spreadsheet_code_json = t.Unicode('').tag(sync=True)
    column_filters_json = t.Unicode('').tag(sync=True)
    column_type_json = t.Unicode('').tag(sync=True)
    has_rendered = t.Bool(True).tag(sync=True)
    user_email = t.Unicode('').tag(sync=True)
    step_summaries_list_json = t.Unicode('').tag(sync=True)
    should_upgrade_mitosheet = t.Bool(False).tag(sync=True)
    data_type_in_mito = t.Unicode('').tag(sync=True)
    received_tours = t.Unicode('').tag(sync=True)
    num_feedbacks = t.Int(0).tag(sync=True)
    num_usages = t.Int(0).tag(sync=True)

    def __init__(self, *args, **kwargs):
        """
        Takes a list of dataframes and strings that are paths to CSV files
        passed through *args.
        """
        # Call the DOMWidget constructor to set up the widget properly
        super(MitoWidget, self).__init__()

        # Set if this is a local deployment
        self.is_local_deployment = is_local_deployment()

        # Mark if it is time for the user to update
        self.should_upgrade_mitosheet = should_upgrade_mitosheet()
            
        # Set up the state container to hold private widget state
        self.steps_manager = StepsManager(args)
        
        # When the widget is first created, it has not been rendered on the frontend yet,
        # but after it is rendered once, this is set to False. This helps us detect 
        # when we are rendering for the first time vs. refreshing the sheet
        self.has_rendered = False

        # Set up starting shared state variables
        self.update_shared_state_variables()

        # Set up message handler
        self.on_msg(self.receive_message)

        # And the api
        self.api = API(self.steps_manager, self.send)

        # We also update the variables that mark how many times
        # the user has given feedback, as well as how many times
        # the user has used the tool
        self.num_feedbacks = len(get_user_field(UJ_FEEDBACKS))
        self.num_usages = len(get_user_field(UJ_MITOSHEET_LAST_FIFTY_USAGES))


    def update_shared_state_variables(self):
        """
        Helper function for updating all the variables that are shared
        between the backend and the frontend through trailets.
        """
        self.sheet_json = self.steps_manager.sheet_json
        self.curr_step_idx = self.steps_manager.curr_step_idx
        self.column_spreadsheet_code_json = self.steps_manager.column_spreadsheet_code_json
        self.code_json = self.steps_manager.code_json
        self.df_names_json = self.steps_manager.df_names_json
        self.df_sources_json = self.steps_manager.df_sources_json
        self.df_shape_json = self.steps_manager.df_shape_json
        self.analysis_name = self.steps_manager.analysis_name
        self.saved_analysis_names_json = saved_analysis_names_json()
        self.column_filters_json = self.steps_manager.column_filters_json
        self.column_type_json = self.steps_manager.column_type_json
        self.user_email = self.steps_manager.user_email
        self.step_summaries_list_json = self.steps_manager.step_summaries_list_json
        self.data_type_in_mito = str(self.steps_manager.data_type_in_mito)
        self.received_tours = self.steps_manager.received_tours


    def handle_edit_event(self, event):
        """
        Handles an edit_event. Per the spec, an edit_event
        updates both the sheet and the codeblock, and as such
        the sheet is re-evaluated and the code for the codeblock
        is re-transpiled.

        Useful for any event that changes the state of both the sheet
        and the codeblock!
        """
        # First, we send this new edit to the evaluator
        self.steps_manager.handle_edit_event(event)

        # We update the state variables 
        self.update_shared_state_variables()

        # Also, write the analysis to a file!
        write_analysis(self.steps_manager)

        # Tell the front-end to render the new sheet and new code with an empty
        # response. NOTE: in the future, we can actually send back some data
        # with the response (like an error), to get this response in-place!        
        self.send({
            'event': 'response',
            'id': event['id']
        })



    def handle_update_event(self, event):
        """
        This event is not the user editing the sheet, but rather information
        that has been collected from the frontend (after render) that is being
        passed back.

        For example:
        - Names of the dataframes
        - Name of an existing analysis
        """
        # If this is just a message to notify the backend that we have rendered, set and return
        if event['type'] == 'has_rendered_update':
            self.has_rendered = True
            return

        self.steps_manager.handle_update_event(event)

        # Update all state variables
        self.update_shared_state_variables()

        # Also, write the analysis to a file!
        write_analysis(self.steps_manager)

        # Tell the front-end to render the new sheet and new code with an empty
        # response. NOTE: in the future, we can actually send back some data
        # with the response (like an error), to get this response in-place!
        self.send({
            'event': 'response',
            'id': event['id']
        })

    def receive_message(self, widget, content, buffers=None):
        """
        Handles all incoming messages from the JS widget. There are two main
        types of events:

        1. edit_event: any event that updates the state of the sheet and the
        code block at once. Leads to reevaluation, and a re-transpile.

        2. update_event: any event that isn't caused by an edit, but instead
        other types of new data coming from the frontend (e.g. the df names 
        or some existing steps).

        3. A log_event is just an event that should get logged on the backend.
        """
        event = content

        try:
            if event['event'] == 'edit_event':
                self.handle_edit_event(event)
            elif event['event'] == 'update_event':
                self.handle_update_event(event)
            elif event['event'] == 'api_call':
                self.api.process_new_api_call(event)
                return 
            
            # NOTE: we don't need to case on log_event above because it always gets
            # passed to this function, and thus is logged. However, we do not log
            # api calls, as they are just noise.
            log_event_processed(event, self.steps_manager)
        except EditError as e:
            print(get_recent_traceback_as_list())
            print(e)
            
            # Log processing this event failed
            log_event_processed(event, self.steps_manager, failed=True, edit_error=e)

            # Report it to the user, and then return
            self.send({
                'event': 'edit_error',
                'id': event['id'],
                'type': e.type_,
                'header': e.header,
                'to_fix': e.to_fix
            })
        except:
            print(get_recent_traceback_as_list())
            # We log that processing failed, but have no edit error
            log_event_processed(event, self.steps_manager, failed=True)
            # Report it to the user, and then return
            self.send({
                'event': 'edit_error',
                'id': event['id'],
                'type': 'execution_error',
                'header': 'Execution Error',
                'to_fix': 'Sorry, there was an error during executing this code.'
            })

def sheet(
        *args, 
        saved_analysis_name=None
    ) -> MitoWidget:
    """
    Renders a Mito sheet. If no arguments are passed, renders an empty sheet. Otherwise, renders
    any dataframes that are passed. Errors if any given arguments are not dataframes or paths to
    CSV files that can be read in as dataframes.

    If running this function just prints text that looks like `MitoWidget(...`, then you need to 
    install the JupyterLab extension manager by running:

    jupyter labextension install @jupyter-widgets/jupyterlab-manager@2;

    Run this command in the terminal where you installed Mito. It should take 5-10 minutes to complete.

    Then, restart your JupyterLab instance, and refresh your browser. Mito should now render.

    NOTE: if you have any issues with installation, please book a demo at https://hubs.ly/H0FL1920
    """
    args = list(args)

    # We error if the saved analysis does not exist
    if saved_analysis_name is not None and read_and_upgrade_analysis(saved_analysis_name) is None:
        log(
            'mitosheet_sheet_call_failed', 
            {'error': f'There is no saved analysis with the name {saved_analysis_name}.'}
        )
        raise ValueError(f'There is no saved analysis with the name {saved_analysis_name}.')

    try:
        # We pass in the dataframes directly to the widget
        widget = MitoWidget(*args) 

        # Log they have personal data in the tool if they passed a dataframe
        # that is not tutorial data or sample data from import docs
        if widget.steps_manager.data_type_in_mito == DataTypeInMito.PERSONAL:
            log('used_personal_data') 

    except:
        log_recent_error('mitosheet_sheet_call_failed')
        raise

    # Then, we log that the call was successful, along with all of it's params
    log(
        'mitosheet_sheet_call',
        dict(
            **{
                'param_saved_analysis_name': saved_analysis_name
            },
            **{
                # NOTE: analysis name is the UUID that mito saves the analysis under
                'steps_manager_analysis_name': widget.steps_manager.analysis_name,
            },
            **{
                'param_num_args': len(args),
                'params_num_str_args': len([arg for arg in args if isinstance(arg, str)]),
                'params_num_df_args': len([arg for arg in args if isinstance(arg, pd.DataFrame)]),
                'params_df_index_type': [str(type(arg.index)) for arg in args if isinstance(arg, pd.DataFrame)],
            }
        )
    )
    
    # If an analysis is passed, we pass this to the widget
    if saved_analysis_name is not None:
        widget.receive_message(widget, {
            'event': 'update_event',
            'id': get_new_id(),
            'type': 'replay_analysis_update',
            'analysis_name': saved_analysis_name,
            'import_summaries': None,
            'clear_existing_analysis': False
        })

    return widget



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import pandas as pd
import pytest

from mitosheet.utils import get_new_id
from mitosheet.errors import EditError
from mitosheet.steps_manager import StepsManager
from mitosheet.tests.test_utils import create_mito_wrapper


def test_create_steps_manager():
    df1 = pd.DataFrame(data={'A': [123]})
    df2 = pd.DataFrame(data={'A': [123]})

    steps_manager = StepsManager([df1, df2])
    assert steps_manager.curr_step_idx == 0
    assert steps_manager.curr_step.step_type == 'initialize'
    assert steps_manager.curr_step.column_metatype == [{'A': 'value'}, {'A': 'value'}]
    assert steps_manager.curr_step.column_spreadsheet_code == [{'A': ''}, {'A': ''}]
    assert steps_manager.curr_step.column_python_code == [{'A': ''}, {'A': ''}]
    assert steps_manager.curr_step.column_evaluation_graph == [{'A': set()}, {'A': set()}]
    assert steps_manager.curr_step.column_python_code == [{'A': ''}, {'A': ''}]
    assert steps_manager.curr_step.dfs[0].equals(df1)
    assert steps_manager.curr_step.dfs[1].equals(df2)

# We assume only column A exists
CELL_EDIT_ERRORS = [
    ('=HI()', 'unsupported_function_error'),
    ('=UPPER(HI())', 'unsupported_function_error'),
    ('=UPPER(HI())', 'unsupported_function_error'),
    ('=C', 'no_column_error'),
    ('=C + D', 'no_column_error'),
    ('=ABCDEFG', 'no_column_error'),
    ('=UPPER(C)', 'no_column_error'),
    ('=B', 'circular_reference_error'),
    ('=UPPER(A, 100)', 'execution_error'),
    ('=UPPER(LOWER(A, 100))', 'execution_error')
]

@pytest.mark.parametrize("formula,error_type", CELL_EDIT_ERRORS)
def test_steps_manager_cell_edit_errors(formula,error_type):
    mito = create_mito_wrapper([123])
    mito.add_column(0, 'B')
    with pytest.raises(EditError) as e_info:
        mito.mito_widget.steps_manager.handle_edit_event({
            'event': 'edit_event',
            'id': get_new_id(),
            'type': 'set_column_formula_edit',
            'step_id': get_new_id(),
            'params': {
                'sheet_index': 0,
                'column_header': 'B',
                'old_formula': '=0',
                'new_formula': formula
            }
        })
    assert e_info.value.type_ == error_type

def test_wrong_column_metatype():
    mito = create_mito_wrapper([123])
    mito.add_column(0, 'B')
    with pytest.raises(EditError) as e_info:
        mito.mito_widget.steps_manager.handle_edit_event({
            'event': 'edit_event',
            'id': get_new_id(),
            'type': 'set_column_formula_edit',
            'step_id': get_new_id(),
            'params': {
                'sheet_index': 0,
                'column_header': 'A',
                'old_formula': '=0',
                'new_formula': '=1'
            }
        })
    assert e_info.value.type_ == 'wrong_column_metatype_error'


def test_overwrites_step_valid():
    mito = create_mito_wrapper([1, 2, 3])
    mito.add_column(0, 'B')

    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'edit_event',
        'id': get_new_id(),
        'type': 'add_column_edit',
        'step_id': mito.mito_widget.steps_manager.curr_step.step_id,
        'params': {
            'sheet_index': 0,
            'column_header': 'C',
            'column_header_index': 1
        }
    })

    steps_manager = mito.mito_widget.steps_manager
    assert len(steps_manager.steps) == 3
    assert steps_manager.dfs[0].equals(pd.DataFrame(data={'A': [1, 2, 3], 'C': [0, 0, 0]}))


def test_failed_overwrite_rolls_back_to_previous_state():
    mito = create_mito_wrapper([1, 2, 3])
    mito.add_column(0, 'B')

    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'edit_event',
        'id': get_new_id(),
        'type': 'add_column_edit',
        'step_id': mito.mito_widget.steps_manager.curr_step.step_id,
        'params': {
            'sheet_index': 1,
            'column_header': 'C',
            'column_header_index': 2
        }
    })

    steps_manager = mito.mito_widget.steps_manager
    assert len(steps_manager.steps) == 2
    assert steps_manager.dfs[0].equals(pd.DataFrame(data={'A': [1, 2, 3], 'B': [0, 0, 0]}))





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.


from mitosheet.step_performers.import_steps.simple_import import SimpleImportStepPerformer


CLEAR_EVENT = 'clear'
CLEAR_PARAMS = []

def execute_clear_update(steps_manager):
    steps_manager.execute_clear()

CLEAR_UPDATE = {
    'event_type': CLEAR_EVENT,
    'params': CLEAR_PARAMS,
    'execute': execute_clear_update
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Exports the transpile function, which takes the backend widget
container and generates transpiled Python code.
"""

from mitosheet.preprocessing import PREPROCESS_STEPS


IN_PREVIOUS_STEP_COMMENT = '# You\'re viewing a previous step. Click fast forward in the Mitosheet above to see the full analysis.'

def transpile(steps_manager, add_comments=True):
    """
    Transpiles the code from the current steps in the steps_manager, 
    displaying up to the checked out step.

    If add_comments, then adds descriptive comments using the step 
    describe functions. 
    """

    code = []

    # First, we transpile all the preprocessing steps
    for PREPROCESS_STEP in PREPROCESS_STEPS:
        preprocess_code = PREPROCESS_STEP['transpile'](steps_manager)
        if len(preprocess_code) > 0:
            code.extend(preprocess_code)

    from mitosheet.steps_manager import get_step_indexes_to_skip
    step_indexes_to_skip = get_step_indexes_to_skip(steps_manager.steps)

    # We only transpile up to the currently checked out step
    for step_index, step in enumerate(steps_manager.steps[:steps_manager.curr_step_idx + 1]):
        # Skip the initalize step, or any step we should skip
        if step.step_type == 'initialize' or step_index in step_indexes_to_skip:
            continue

        # The total code for this step
        step_code = []

        # NOTE: we add a comment if add_comment is True
        if add_comments:
            step_code.append(
                '# ' + step.step_performer.describe(
                    **step.params,
                    df_names=step.df_names
                )
            )
        
        transpiled_code = step.step_performer.transpile(
            step.prev_state,
            step.post_state,
            step.execution_data,
            **step.params
        )
        
        # Make sure we don't add comments for steps with no code
        if len(transpiled_code) > 0:

            step_code.extend(
                transpiled_code
            )
            
            code.extend(step_code)

    # If we have a historical step checked out, then we add a comment letting
    # the user know this is the case
    if steps_manager.curr_step_idx != len(steps_manager.steps) - 1:
        code.append(IN_PREVIOUS_STEP_COMMENT)

    return {
        'imports': f'from mitosheet import *',
        'code': code
    }


import os

def get_path_join(event):
    """
    Takes an event with path parts in a list, turns them into
    an actual path, normalizes it, and sends it back.

    Making path joining robust / cross platform compatible is actually really 
    tough, which is why we rely on Python to do it. We could use a JS package
    but instead of adding another dependency that might be incompatible, we just
    use the Python code on the backend.
    """ 
    path_parts = event['path_parts']

    # Join the path and normalize it
    path = os.path.join(*path_parts)
    path = os.path.normpath(path)

    if path == '.':
        path = os.getcwd()

    return path
    
    
    


import json
from mitosheet.errors import get_recent_traceback
from mitosheet.api.graph.box import get_box_code, get_box_plot
from mitosheet.api.graph.bar import get_bar_chart, get_bar_code
from mitosheet.api.graph.scatter import get_scatter_code, get_scatter_plot
from mitosheet.api.graph.histogram import get_histogram, get_histogram_code
import pandas as pd
import plotly.express as px

from mitosheet.mito_analytics import log, log_recent_error
from mitosheet.sheet_functions.types.utils import get_mito_type, NUMBER_SERIES

from mitosheet.api.graph.graph_utils import (
    MAX_UNIQUE_NON_NUMBER_VALUES, MAX_UNIQUE_NON_NUMBER_VALUES_COMMENT, X, Y,
    HISTOGRAM, SCATTER, BAR, HISTOGRAM, BOX, SUMMARY_STAT,
    filter_df_to_top_unique_values_in_series, get_html_and_script_from_figure
)


def get_column_summary_graph(axis, df, axis_data_array):
    """
    One Axis Graphs heuristics:
    1. Number Column - we do not filtering. These graphs are pretty efficient up to 1M rows
    2. Non-number column. We filter to the top 10k values, as the graphs get pretty laggy 
       beyond that
    """
    column_header: str = axis_data_array[0]
    series: pd.Series = df[column_header]
    mito_type = get_mito_type(series)

    graph_title = f'{column_header} Frequencies'

    filtered = False
    if mito_type != NUMBER_SERIES:
        if series.nunique() > MAX_UNIQUE_NON_NUMBER_VALUES:
            title = f'{graph_title} {MAX_UNIQUE_NON_NUMBER_VALUES_COMMENT}'
            df = filter_df_to_top_unique_values_in_series(
                df, 
                series, 
                MAX_UNIQUE_NON_NUMBER_VALUES
            )
            # Set series as the newly filtered series
            series = df[column_header]

            filtered = True

        # Fill NaN values with 'NaN' so they are displayed in the graph.
        series = series.fillna('NaN')
            
    labels = {axis: ''}

    kwargs = {
        axis: series,
        'labels': labels,
        'title': graph_title,
    }

    fig = px.histogram(
        **kwargs        
    )

    log(f'generate_column_summary_stat_graph', {
        f'param_is_number_series_{axis}': mito_type == NUMBER_SERIES,
        'param_filtered': filtered
    })

    # Because the Column Summary Stat Graph is never zoomed in 
    # we always return False as the second parameter
    return fig, False

def get_graph(event, steps_manager):
    """
    Creates a graph of the passed parameters, and sends it back as a PNG
    string to the frontend for display.

    Params:
    - graph_type
    - sheet_index
    - x_axis_column_header_array (optional)
    - y_axis_column_header_array (optional)
    - height (optional) - int representing the div width
    - width (optional) - int representing the div width

    If only an x axis is given, and if the series is a numeric series,
    will return a histogram. Otherwise, as long as there are less than 
    20 distinct items in the series, will return a bar chart of the 
    value count. Otherwise, will return nothing.
    """
    keys = event.keys()

    # Get graph type 
    graph_type = event['graph_type'] if 'graph_type' in keys else None

    # Get the x axis params, if they were provided
    x_axis_column_headers = event['x_axis_column_headers'] if event['x_axis_column_headers'] is not None else []
    x_axis = len(x_axis_column_headers) > 0

    # Get the y axis params, if they were provided
    y_axis_column_headers = event['y_axis_column_headers'] if event['y_axis_column_headers'] is not None else []
    y_axis = len(y_axis_column_headers) > 0
    
    # Find the height and the width, defaulting to fill whatever container its in
    height = event["height"] if 'height' in keys else '100%'
    width = event["width"] if 'width' in keys else '100%'

    sheet_index = event["sheet_index"] if "sheet_index" in keys else None

    try:
        # First, handle edge cases
        if not x_axis and not y_axis:
            # If no axes provided, return
            log(f'failed_generate_graph', {
                'params_graph_type': graph_type,
                'params_failed_reason': 'no axis data'
            })

            return ''
        if sheet_index is None or graph_type is None:
            # If no sheet_index or graph type is provided, return
            failed_reason = 'no selected_sheet_index' if sheet_index is None else 'no graph_type'
            log(f'failed_generate_graph', {
                'params_graph_type': graph_type,
                'params_failed_reason': failed_reason
            })
            return ''
        if x_axis and len(x_axis_column_headers) > 1 and y_axis and len(y_axis_column_headers) > 1:
            # If both axises have more than 1 series, return
            log(f'failed_generate_graph', {
                'params_graph_type': graph_type,
                'params_failed_reason': 'both axises have multiple columns'
            })
            return ''

        # Create a copy of the dataframe, just for safety.
        df: pd.DataFrame = steps_manager.dfs[sheet_index].copy()
        df_name: str = steps_manager.curr_step.df_names[sheet_index]

        # Handle the graphs in alphabetical order
        if graph_type == BAR:
            fig, is_zoomed = get_bar_chart(df, x_axis_column_headers, y_axis_column_headers)
            generation_code = get_bar_code(df, x_axis_column_headers, y_axis_column_headers, df_name)
        elif graph_type == BOX:
            # Box plots are only defined on one axis. The UI should enforce that 
            # it is never the case that a box plot is selected with series for both 
            # the x and y axis. However, if it does happen, we default to taking the x axis. 
            if x_axis:
                fig, is_zoomed = get_box_plot(X, df, x_axis_column_headers)
                generation_code = get_box_code(X, df, x_axis_column_headers, df_name)
            else:
                fig, is_zoomed = get_box_plot(Y, df, y_axis_column_headers)
                generation_code = get_box_code(Y, df, y_axis_column_headers, df_name)
        elif graph_type == HISTOGRAM:
            # Histograms are only defined on one axis. The UI should enforce that 
            # it is never the case that a histogram is selected with series for both 
            # the x and y axis. However, if it does happen, we default to taking the x axis. 
            if x_axis:
                fig, is_zoomed = get_histogram(X, df, x_axis_column_headers)
                generation_code = get_histogram_code(X, df, x_axis_column_headers, df_name)
            else:
                fig, is_zoomed = get_histogram(Y, df, y_axis_column_headers)
                generation_code = get_histogram_code(Y, df, y_axis_column_headers, df_name)
        elif graph_type == SCATTER:
            fig, is_zoomed = get_scatter_plot(df, x_axis_column_headers, y_axis_column_headers)
            generation_code = get_scatter_code(df, x_axis_column_headers, y_axis_column_headers, df_name)
        elif graph_type == SUMMARY_STAT:
            # We handle summary stats separately from the histogram, for now, because 
            # we only let the user use a histogram with all numeric data, whereas the column
            # summary stats may not be all numeric data. 
            fig, is_zoomed = get_column_summary_graph(X, df, x_axis_column_headers)
            generation_code = ''
            
        # 1) Get rid of some of the default white space
        # 2) Add a range slider
        # 3) Set the colors of the graphs so they stand out more
        fig.update_layout(
            margin=dict(
                l=0,
                r=0,
                t=30, 
                b=30,
            ),
            xaxis=dict(
                rangeslider=dict(
                    visible=True
                ),
            ),
        )

        # Make the rangeslider take up 5% of the height of the plot space
        fig.update_xaxes(
            rangeslider_thickness = 0.05
        )

        return_object = get_html_and_script_from_figure(
            fig, height, width
        )

        # Tell the frontend if the graph is zoomed in
        return_object['is_zoomed'] = is_zoomed

        return_object['generation_code'] = generation_code

        return json.dumps(return_object),

    except Exception as e:
        print(get_recent_traceback())
        log_recent_error('graph_generation_errored')
        # As not being able to make a graph is a non-critical error that doesn't
        # result from user interaction, we don't want to throw an error if something
        # weird happens, so we just return nothing in this case
        return ''


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains tests to make sure that the mito analytics test is
performing correctly
"""

# TODO: add tests for anonymization


"""
Mito currently collects logs on basic user interactions with the app, 
so that we can improve the tool.

However, _we collect no logs that contain personal user data_. No private
data is taken from the users computer, as can be verified by this app.

Our general approach to logging can be understood as:
0. Log no information about the data users use in Mito. Not a drop!
1. Users are identified by a single random ID that we generate on the client-side, 
   here. We store this in the ~/.mito/user.json file, and we use it as the permanent
   ID for all users.
2. All logging is on the backend. This avoids us worrying about blockers or needing
   to associate with ad-tech at all.
3. We generate a single log event for a single action taken by the user. That means that if
   the user takes an action that causes an error, the error is logged _with_ that action.

   This appears to be good practice, as it allows us to associate what actions are taken
   with their result very effectively!
"""
from mitosheet.user.schemas import UJ_MITOSHEET_TELEMETRY
import platform
import sys
import subprocess

from mitosheet.parser import parse_formula

import analytics
# Write key taken from segement.com
analytics.write_key = '6I7ptc5wcIGC4WZ0N1t0NXvvAbjRGUgX' 

from mitosheet.user import (
    UJ_INTENDED_BEHAVIOR, UJ_STATIC_USER_ID, UJ_USER_SALT, UJ_USER_EMAIL, 
    UJ_FEEDBACKS, 
)
from mitosheet.user import get_user_field, is_local_deployment, is_running_test, is_local_deployment
from mitosheet._version import __version__
from mitosheet.errors import get_recent_traceback, get_recent_traceback_as_list

# We cache if logging is turned on to make sure we dont
# make unecessary calls to read the user.json
TELEMETRY_TURNED_ON = None

def telemetry_turned_on():
    """
    Helper function that tells if you if logging is turned on or
    turned off on the entire Mito instance
    """
    global TELEMETRY_TURNED_ON
    if TELEMETRY_TURNED_ON is None:
        TELEMETRY_TURNED_ON = get_user_field(UJ_MITOSHEET_TELEMETRY)
    
    return TELEMETRY_TURNED_ON

# When we anonymize, we use some combination of these words
# to construct new private words
valid_words = ['cat', 'dog', 'hat', 'time', 'person', 'year', 'way', 'thing', 'man', 'world', 'life', 'born', 'part', 'child', 'eye', 'woman', 'place', 'work', 'fall', 'case', 'point', 'company', 'number', 'group', 'problem', 'fact']

# We use the same salt to anonymize_words, and we read
# this salt in once the function is called for the first
# time, to make sure it's initialized properly
salt = None
def anonymize_word(word):
    """
    Helper function that turns a column header into
    a totally anonymous version of the column header,
    as to not leak _any_ user data
    """
    # We make sure that the salt is read in after the entire
    # app has been initalized, so that we don't have to read
    # from the file all the time
    global salt
    if salt is None:
        salt = get_user_field(UJ_USER_SALT)

    word = str(word)

    # We select three indexes from the valid_words list, and concatenate them
    index_one = int(hash(salt + word + '0')) % len(valid_words)
    index_two = int(hash(salt + word + '1')) % len(valid_words)
    index_three = int(hash(salt + word + '2')) % len(valid_words)

    return valid_words[index_one] + valid_words[index_two] + valid_words[index_three]


def anonymize_formula(formula):
    """
    Helper function that anonymizes formula to 
    make sure that no private data is included in it.
    """
    # We just input a random address, as we don't use it
    _, _, dependencies = parse_formula(formula, 'A', throw_errors=False)
    
    for dependency in dependencies:
        formula = formula.replace(dependency, anonymize_word(dependency))
    
    return formula

def get_jupyterlab_version():
    try:
        from jupyterlab import __version__ as jupyterlab_version
    except:
        jupyterlab_version = 'No JupyterLab'

    return jupyterlab_version

def get_jupyter_labextension_list():
    """
    This gets the current output from the command `jupyter labextension list`
    which is very helpful in debugging the users installation.


    Because this function can be _very_ expensive to run for some users, we should
    only use this sparingly and not during every log event!

    Further note: this function is non critical, so we do _not_ throw errors
    when this fails, and instead just log the error and continue.
    """
    try:
        completed_process_mitosheet3 = subprocess.run([sys.executable, '-m', 'jupyter', 'labextension', 'list'], capture_output=True)
        all_output = completed_process_mitosheet3.stdout.decode() + completed_process_mitosheet3.stderr.decode()
        return all_output.splitlines()
    except Exception as e:
        log_recent_error('get_jupyerlab_extension_list_failed')
        return []

def get_installed_mitosheets_pip_show():
    """
    This gets the version of mitosheet and mitosheet3 that are installed
    through pip, which allows us to see if the installed version of mitosheet
    is actually what is running. This allows us to determine if the user has
    failed to refresh their kernel!

    Because this function can be expensive to run for some users, we should
    only use this sparingly and not during every log event!

    Further note: this function is non critical, so we do _not_ throw errors
    when this fails, and instead just log the error and continue.
    """
    try:
        completed_process_mitosheet3 = subprocess.run([sys.executable, '-m', 'pip', 'show', 'mitosheet3'], capture_output=True)
        all_output = completed_process_mitosheet3.stdout.decode() + completed_process_mitosheet3.stderr.decode()
        
        completed_process_mitosheet = subprocess.run([sys.executable, '-m', 'pip', 'show', 'mitosheet'], capture_output=True)
        all_output += completed_process_mitosheet.stdout.decode() + completed_process_mitosheet.stderr.decode()
        return all_output.splitlines()
    except Exception as e:
        log_recent_error('get_installed_mitosheets_pip_show_failed')
        return []


def log(log_event, params=None, check_mitosheet_install=False):
    """
    Helper function that logs an event with the given parameters. However,
    this function is also responsible for making sure _zero_ private data
    leaves the users computer.

    To accomplish this, we anonymize all data. To do so, we hash this data
    with a private secret that only the user has, making it impossible
    for us to brute force it.

    NOTE: check_mitosheet_install can be an expensive operation for some
    users, so we should try and minimize where / how we use it!
    """
    if not telemetry_turned_on():
        return

    if params is None:
        params = {}
    
    # Add the python properties to every log event we can
    python_properties = {
        'version_python': sys.version_info,
        'version_jupyterlab': get_jupyterlab_version(),
        'version_mito': __version__ 
    }
    
    if check_mitosheet_install:
        pip_shows = get_installed_mitosheets_pip_show()
        jupyter_labextension_list = get_jupyter_labextension_list()
    else:
        pip_shows = []
        jupyter_labextension_list = []
    
    params = dict(
        **params,
        **python_properties,
        **{
            'pip_shows': pip_shows,
            'jupyter_labextension_list': jupyter_labextension_list
        }
    )

    # Private params are where we _make sure_ that no private
    # user data leaves the users machine. We replace any potentially
    # non-private params with private versions of them
    private_params = {}
    try:
        for key, value in params.items():
            # We take any of the items that contain private user data, and we anonymize them
            if ('column_header' in key \
                or 'df_name' in key \
                or 'dataframe_name' in key \
                or 'analysis_name' in key \
                or 'merge_key' in key ) \
                and 'index' not in key:
                private_params[key] = anonymize_word(value)
            elif 'formula' in key:
                # We make sure to remove any private references from a formula
                private_params[key] = anonymize_formula(value)
            elif 'file_names' in key:
                # Just count the number of files, instead of keeping their names
                private_params[key] = len(value)
            elif 'rows' in key or 'columns' in key or 'selected_columns' in key:
                # We anonymize all of the column headers in any list
                private_params[key] = [anonymize_word(v) for v in value]
            elif 'values' in key:
                # for pivot params, we anonymize the column headers
                # and keep the agg functions
                private_params[key] = {anonymize_word(k): v for k, v in value.items()}
            elif 'filters' in key:
                # For filters, we get rid of the values, but keep the conditions
                # as well as if a value was put in
                filter_log = []

                for filter_or_group in value:
                    if 'filters' in filter_or_group:
                        filter_log.append([{'condition': filter_['condition'],'value': len(str(filter_['value']))} for filter_ in filter_or_group['filters']])
                    else:
                        filter_log.append({'condition': filter_or_group['condition'],'value': len(str(filter_or_group['value']))})

                private_params[key] = filter_log
            elif 'import_summaries' in key:
                if value is not None:
                    private_params[key] = len(value)
                else:
                    private_params[key] = value

            else:
                private_params[key] = value
    except:
        # We log if we fail to build the private logs, so that we 
        # know that we have some issues here
        private_params['failed_building'] = True

    # NOTE: we do not log anything when tests are running
    if not is_running_test():
        analytics.track(
            get_user_field(UJ_STATIC_USER_ID), 
            log_event, 
            private_params
        )


def identify():
    """
    Helper function for identifying a user. We just take
    their python version, mito version, and email.
    """
    if not telemetry_turned_on():
        return

    static_user_id = get_user_field(UJ_STATIC_USER_ID)
    user_email = get_user_field(UJ_USER_EMAIL)
    intended_behavior = get_user_field(UJ_INTENDED_BEHAVIOR)
    feedbacks = get_user_field(UJ_FEEDBACKS)
    local = is_local_deployment()
    jupyterlab_version = get_jupyterlab_version()
    operating_system = platform.system()

    
    if not is_running_test():
        # NOTE: we do not log anything when tests are running
        analytics.identify(static_user_id, {
            'version_python': sys.version_info,
            'version_mito': __version__,
            'version_jupyterlab': jupyterlab_version,
            'operating_system': operating_system,
            'email': user_email,
            'local': local,
            UJ_INTENDED_BEHAVIOR: intended_behavior,
            UJ_FEEDBACKS: feedbacks
        })


def log_recent_error(log_event=None):
    """
    A helper function for logging the most recent error that has occured.

    log_event defaults to an execution_error_log_event
    """
    if log_event is None:
        log_event = 'execution_error_log_event'

    # We get the error, see more here: https://wiki.python.org/moin/HandlingExceptions
    e = sys.exc_info()[0]

    # If we have some other error, we just report this as an execution error
    log(
        log_event, 
        {
            'header': 'Execution Error', 
            'to_fix': 'Sorry, there was an error during executing this code.',
            'error': str(e),
            'traceback': get_recent_traceback_as_list()
        }
    )


def log_event_processed(event, steps_manager, failed=False, edit_error=None):
    """
    Helper function for logging when an event is processed
    by the widget state container. 

    Does it's best to fill in helpful meta-data for interpreting the event
    including the state of the steps_manager _after_ the step
    was applied.

    NOTE: if processing the event fails, then failed should be True. If there was an
    edit error that was thrown during the processing of the event, then edit_error
    should be set to that error.
    """
    try:
        # First, we get all the params of the event, and append them with _params_
        if 'params' in event:
            # If this is an edit event, then, we get the params from the params
            # key all at once
            event_properties = {
                'params_' + key: value for key, value in event['params'].items()
            }
        else:
            # Otherwise, we get the params from the highest level of the event 
            event_properties = {
                'params_' + key: value for key, value in event.items()
            }

        # We also get some metadata about the widget state container at this state
        # NOTE: we keep this as underscored with wsc for backwards compatibility with
        # our logs!
        steps_manager_properties = {
            # NOTE: analysis name is the UUID that mito saves the analysis under
            'wsc_analysis_name': steps_manager.analysis_name,
            # NOTE: change this when code fixing this logic is merged in
            'wsc_data_type_in_mito': str(steps_manager.data_type_in_mito),
            'wsc_local': is_local_deployment(),
            'wsc_curr_step_idx': steps_manager.curr_step_idx,
            'wsc_curr_step_type': steps_manager.curr_step.step_type,
        }

        if event['type'] == 'replay_analysis_update' or event['type'] == 'save_analysis_update':
            # We log the steps, when the user saves or replays an analysis
            steps_manager_properties['steps_manager_steps'] = [
                step.step_type for step in steps_manager.steps
            ]

        # We also check there is an edit_error, and if there is, then we add the error logs
        if edit_error is not None:
            recent_traceback = get_recent_traceback_as_list()
            error_properties = {
                'error_type': edit_error.type_,
                'error_header': edit_error.header,
                'error_to_fix': edit_error.to_fix,
                'error_traceback': recent_traceback,
                # We get the last line of the error as it makes it much easier
                # to easily analyze on error messages 
                'error_traceback_last_line': recent_traceback[-1],
            }
        elif failed:
            # Otherwise, if there is no edit_error, and we still failed, then we must have
            # gotten an execution error
            recent_traceback = get_recent_traceback_as_list()
            error_properties = {
                'error_type': 'execution_error',
                'error_header': 'Execution Error',
                'error_traceback': recent_traceback,
                # We get the last line of the error as it makes it much easier
                # to easily analyze on error messages
                'error_traceback_last_line': recent_traceback[-1],
            }
        else:
            error_properties = {}

        # We choose to log the event type, as it is the best high-level item for our logs
        # and we append a _failed if the event failed in doing this.
        log_event = event['type'] + ('_failed' if failed else '')

        log(
            log_event, 
            dict(
                **event_properties,
                **steps_manager_properties,
                **error_properties
            )
        )
    except:
        # We don't want logging to ever brick the application, so if the logging fails
        # we just log simple information about the event. This should never occur, but it
        # is just a precaution - some defensive programming so to speak
        log(
            event['type'],
        )



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import pytest
import os

from mitosheet.saved_analyses import _get_all_analysis_filenames, _delete_analyses


@pytest.fixture(scope="session", autouse=True)
def cleanup_files():
    """
    This fixture is responsible for cleaning up all saved analyses that result from
    test runs. 

    It does so by reading in all analyses at the start of the testing session, 
    and then removing any new analyses at the end. 

    NOTE: if you use the app while the tests are running, it might delete
    your analyses - but this is really rare!
    """
    # This test runs before 
    old_analysis_filenames = _get_all_analysis_filenames()

    yield # All tests run here

    # This code runs after all the tests
    curr_analysis_filenames = _get_all_analysis_filenames()
    # Find all the new analysis file names (generated by tests)
    new_analysis_filenames = curr_analysis_filenames.difference(old_analysis_filenames)
    # Delete them
    _delete_analyses(new_analysis_filenames)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains utilities used in multiple sheet functions.
"""
import pandas as pd
import numpy as np

def fill_series_to_length(series, length):
    """
    Extends a given series to length, padding it with the last
    defined element in the series. 
    Given all dataframes we use are square, this only occurs when one adds
    1 + df['nums'] (where df['nums'] has more than one row), and then our 
    as_type code converts this to:
    pd.Series(data=[1]) + pd.Series(data=[1, 2]) = pd.Series(data=[1, NaN])
    If we ever are adding non-equal length series, we need to fix this!
    """

    # if the series longer than 1, it must be the correct length because 
    # pandas makes dfs square. therefore, filling the series overwrites NaN's
    if len(series) > 1:
        return series

    return series.append(
        pd.Series(data=([None] * (length - len(series)))), ignore_index=True
    ).ffill()

def fill_series_list_to_max(series_list):
    """
    Extends all series in the series_list to the max_length of any series
    in the list, padding them all with their last defined elements. 
    Useful so we can add two series together, as:
    -----
    pd.Series(data=[1]) + pd.Series(data=[1, 2]) = pd.Series(data=[1, Nan])
    -----
    which isn't really what we want!
    """
    max_length = max([series.size for series in series_list])
    return [fill_series_to_length(series, max_length) for series in series_list]